import pytest

import geomstats.backend as gs
from geomstats.test.random import (
    EmbeddedSpaceRandomDataGenerator,
    MatrixVectorSpaceRandomDataGenerator,
    VectorSpaceRandomDataGenerator,
)
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.complex_manifold import ComplexManifoldTestCase
from geomstats.test_cases.geometry.manifold import ManifoldTestCase
from geomstats.test_cases.geometry.mixins import ProjectionTestCaseMixins
from geomstats.vectorization import get_batch_shape

# TODO: vec with tangent_vecs may not be being tested sufficiently well
# i.e. tests may pass but just because it is the repetition of points

# TODO: define better where to use pytes.mark.mathprop

# TODO: enumerate tests for which random is not enough

# TODO: review uses of gs.ones and gs.zeros

# TODO: review "passes" in tests (maybe not implemented error?)


class _VectorSpaceTestCaseMixins(ProjectionTestCaseMixins):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = VectorSpaceRandomDataGenerator(self.space)
        super().setup_method()

    @pytest.mark.random
    def test_random_point_is_tangent(self, n_points, atol):
        """Check random point is tangent.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        points = self.data_generator.random_point(n_points)

        res = self.space.is_tangent(points, base_point=None, atol=atol)
        self.assertAllEqual(res, gs.ones(n_points, dtype=bool))

    @pytest.mark.random
    def test_to_tangent_is_projection(self, n_points, atol):
        """Check to_tangent is same as projection.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        vec = self.data_generator.random_point(n_points)
        result = self.space.to_tangent(vec)
        expected = self.space.projection(vec)

        self.assertAllClose(result, expected, atol=atol)


class VectorSpaceTestCase(_VectorSpaceTestCaseMixins, ManifoldTestCase):
    @pytest.mark.mathprop
    def test_basis_cardinality(self):
        """Check number of basis elements is the dimension.

        Parameters
        ----------
        atol : float
            Absolute tolerance.
        """
        basis = self.space.basis
        self.assertEqual(basis.shape[0], self.space.dim)

    @pytest.mark.mathprop
    def test_basis_belongs(self, atol):
        """Check basis elements belong to vector space.

        Parameters
        ----------
        atol : float
            Absolute tolerance.
        """
        result = self.space.belongs(self.space.basis, atol=atol)
        self.assertAllEqual(result, gs.ones_like(result))

    def test_basis(self, expected, atol):
        self.assertAllClose(self.space.basis, expected, atol=atol)


class ComplexVectorSpaceTestCase(_VectorSpaceTestCaseMixins, ComplexManifoldTestCase):
    pass


class MatrixVectorSpaceTestCase(VectorSpaceTestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = MatrixVectorSpaceRandomDataGenerator(self.space)
        super().setup_method()

    def test_basis_representation(self, matrix_representation, expected, atol):
        vec = self.space.basis_representation(matrix_representation)
        self.assertAllClose(vec, expected, atol=atol)

    @pytest.mark.vec
    def test_basis_representation_vec(self, n_reps, atol):
        matrix_representation = self.data_generator.random_point()
        expected = self.space.basis_representation(matrix_representation)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    matrix_representation=matrix_representation,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["matrix_representation"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_basis_representation_and_basis(self, n_points, atol):
        mat = self.data_generator.random_point(n_points)
        vec = self.space.basis_representation(mat)

        res = gs.einsum("...i,...ijk->...jk", vec, self.space.basis)
        self.assertAllClose(res, mat, atol=atol)

    def test_matrix_representation(self, basis_representation, expected, atol):
        mat = self.space.matrix_representation(basis_representation)
        self.assertAllClose(mat, expected, atol=atol)

    @pytest.mark.vec
    def test_matrix_representation_vec(self, n_reps, atol):
        basis_representation = gs.random.rand(self.space.dim)
        expected = self.space.matrix_representation(basis_representation)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    basis_representation=basis_representation,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["basis_representation"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_matrix_representation_belongs(self, n_points, atol):
        vec = gs.reshape(gs.random.rand(n_points * self.space.dim), (n_points, -1))
        point = self.space.matrix_representation(vec)

        self.test_belongs(point, gs.ones(n_points, dtype=bool), atol)

    @pytest.mark.random
    def test_matrix_representation_after_basis_representation(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        vec = self.space.basis_representation(point)
        point_ = self.space.matrix_representation(vec)

        self.assertAllClose(point_, point, atol=atol)

    @pytest.mark.random
    def test_basis_representation_after_matrix_representation(self, n_points, atol):
        vec = gs.reshape(gs.random.rand(n_points * self.space.dim), (n_points, -1))
        point = self.space.matrix_representation(vec)
        vec_ = self.space.basis_representation(point)

        self.assertAllClose(vec_, vec, atol=atol)


class ComplexMatrixVectorSpaceTestCase(ComplexVectorSpaceTestCase):
    pass


class LevelSetTestCase(ProjectionTestCaseMixins, ManifoldTestCase):
    # TODO: need to develop `intrinsic_after_extrinsic` and `extrinsic_after_intrinsic`
    # TODO: class to handle `extrinsinc-intrinsic` mixins?

    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = EmbeddedSpaceRandomDataGenerator(self.space)
        super().setup_method()

    def test_submersion(self, point, expected, atol):
        submersed_point = self.space.submersion(point)
        self.assertAllClose(submersed_point, expected, atol=atol)

    def test_submersion_is_zero(self, point, submersion_shape, atol):
        # TODO: keep?
        batch_shape = get_batch_shape(self.space.point_ndim, point)
        expected = gs.zeros(batch_shape + submersion_shape)

        self.test_submersion(point, expected, atol)

    @pytest.mark.vec
    def test_submersion_vec(self, n_reps, atol):
        # TODO: mark as redundant? belongs suffices?
        point = self.space.embedding_space.random_point()
        expected = self.space.submersion(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_tangent_submersion(self, vector, point, expected, atol):
        submersed_vector = self.space.tangent_submersion(vector, point)
        self.assertAllClose(submersed_vector, expected, atol=atol)

    def test_tangent_submersion_is_zero(
        self, tangent_vector, point, tangent_submersion_shape, atol
    ):
        # TODO: keep?
        batch_shape = get_batch_shape(self.space.point_ndim, tangent_vector, point)
        expected = gs.zeros(batch_shape + tangent_submersion_shape)

        self.test_tangent_submersion(tangent_vector, point, expected, atol)

    @pytest.mark.vec
    def test_tangent_submersion_vec(self, n_reps, atol):
        # TODO: mark as redundant? is_tangent suffices?
        vector, point = self.data_generator.random_point(2)
        expected = self.space.tangent_submersion(vector, point)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    vector=vector,
                    point=point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["vector", "point"],
            n_reps=n_reps,
            vectorization_type="sym" if self.tangent_to_multiple else "repetition-0",
            expected_name="expected",
        )
        self._test_vectorization(vec_data)


class _OpenSetTestCaseMixins(ProjectionTestCaseMixins):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = EmbeddedSpaceRandomDataGenerator(self.space)
        super().setup_method()

    @pytest.mark.random
    def test_to_tangent_is_tangent_in_embedding_space(self, n_points):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        res = self.space.embedding_space.is_tangent(tangent_vec, base_point)

        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)


class OpenSetTestCase(_OpenSetTestCaseMixins, ManifoldTestCase):
    pass


class VectorSpaceOpenSetTestCase(OpenSetTestCase):
    pass


class ComplexVectorSpaceOpenSetTestCase(
    _OpenSetTestCaseMixins, ComplexManifoldTestCase
):
    pass


class ImmersedSetTestCase(ProjectionTestCaseMixins, ManifoldTestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = EmbeddedSpaceRandomDataGenerator(self.space)
        super().setup_method()

    def test_immersion(self, point, expected, atol):
        res = self.space.immersion(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_immersion_belongs(self, n_points, atol):
        point = self.data_generator.random_point(n_points)

        embedded_point = self.space.immersion(point)

        expected = gs.ones(n_points, dtype=bool)
        res = self.space.embedding_space.belongs(embedded_point, atol=atol)
        self.assertAllEqual(res, expected)

    def test_tangent_immersion(self, tangent_vec, base_point, expected, atol):
        res = self.space.tangent_immersion(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_tangent_immersion_is_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        embedded_point = self.space.immersion(base_point)
        embedded_tangent_vec = self.space.tangent_immersion(tangent_vec, base_point)

        expected = gs.ones(n_points, dtype=bool)
        res = self.space.embedding_space.is_tangent(
            embedded_tangent_vec, embedded_point, atol=atol
        )
        self.assertAllEqual(res, expected)

    def test_jacobian_immersion(self, base_point, expected, atol):
        res = self.space.jacobian_immersion(base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_hessian_immersion(self, base_point, expected, atol):
        res = self.space.hessian_immersion(base_point)
        self.assertAllClose(res, expected, atol=atol)


class DiffeomorphicManifoldTestCase(ManifoldTestCase):
    pass



from geomstats.test.test_case import TestCase


class KernelDensityEstimationClassifierTestCase(TestCase):
    def test_predict(self, estimator, X_train, y_train, X_test, y_test):
        estimator.fit(X_train, y_train)

        y_pred = estimator.predict(X_test)
        self.assertAllEqual(y_pred, y_test)

    def test_predict_proba(self, estimator, X_train, y_train, X_test, expected, atol):
        estimator.fit(X_train, y_train)

        proba = estimator.predict_proba(X_test)
        self.assertAllClose(proba, expected, atol=atol)



import pytest

from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase
from geomstats.vectorization import repeat_point

# TODO: more random tests (e.g. inner product)

# TODO: add lie algebra related tests


class _InvariantMetricTestCaseMixins(RiemannianMetricTestCase):
    def test_inner_product_at_identity(
        self, tangent_vec_a, tangent_vec_b, expected, atol
    ):
        res = self.space.metric.inner_product_at_identity(tangent_vec_a, tangent_vec_b)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_inner_product_at_identity_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.inner_product_at_identity(
            tangent_vec_a, tangent_vec_b
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_a=tangent_vec_a,
                    tangent_vec_b=tangent_vec_b,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec_a", "tangent_vec_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_invariance(self, n_points, atol):
        left = self.space.metric.left

        group_elem_1 = self.data_generator.random_point(n_points)
        group_elem_2 = self.data_generator.random_point(n_points)

        tangent_vec_11 = self.data_generator.random_tangent_vec(group_elem_1)
        tangent_vec_12 = self.data_generator.random_tangent_vec(group_elem_1)

        dtrans_2 = self.space.tangent_translation_map(
            group_elem_2, left=left, inverse=False
        )

        if left:
            group_elem_3 = self.space.compose(group_elem_2, group_elem_1)
        else:
            group_elem_3 = self.space.compose(group_elem_1, group_elem_2)

        tangent_vec_31 = dtrans_2(tangent_vec_11)
        tangent_vec_32 = dtrans_2(tangent_vec_12)

        inner_prod = self.space.metric.inner_product(
            tangent_vec_11, tangent_vec_12, group_elem_1
        )
        inner_prod_ = self.space.metric.inner_product(
            tangent_vec_31, tangent_vec_32, group_elem_3
        )
        self.assertAllClose(inner_prod, inner_prod_, atol=atol)


class InvariantMetricMatrixTestCase(_InvariantMetricTestCaseMixins):
    def test_structure_constant(
        self, tangent_vec_a, tangent_vec_b, tangent_vec_c, expected, atol
    ):
        res = self.space.metric.structure_constant(
            tangent_vec_a, tangent_vec_b, tangent_vec_c
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_structure_constant_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_c = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.structure_constant(
            tangent_vec_a, tangent_vec_b, tangent_vec_c
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_a=tangent_vec_a,
                    tangent_vec_b=tangent_vec_b,
                    tangent_vec_c=tangent_vec_c,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec_a", "tangent_vec_b", "tangent_vec_c"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_dual_adjoint(self, tangent_vec_a, tangent_vec_b, expected, atol):
        res = self.space.metric.dual_adjoint(tangent_vec_a, tangent_vec_b)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_dual_adjoint_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.dual_adjoint(tangent_vec_a, tangent_vec_b)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_a=tangent_vec_a,
                    tangent_vec_b=tangent_vec_b,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec_a", "tangent_vec_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_dual_adjoint_structure_constant(self, n_points, atol):
        base_point = repeat_point(self.space.identity, n_reps=n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_c = self.data_generator.random_tangent_vec(base_point)

        result = self.space.metric.inner_product_at_identity(
            self.space.metric.dual_adjoint(tangent_vec_a, tangent_vec_b), tangent_vec_c
        )
        expected = self.space.metric.structure_constant(
            tangent_vec_a, tangent_vec_c, tangent_vec_b
        )
        self.assertAllClose(result, expected, atol=atol)

    def test_connection_at_identity(self, tangent_vec_a, tangent_vec_b, expected, atol):
        res = self.space.metric.connection_at_identity(tangent_vec_a, tangent_vec_b)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_connection_at_identity_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)
        expected = self.space.metric.connection_at_identity(
            tangent_vec_a, tangent_vec_b
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_a=tangent_vec_a,
                    tangent_vec_b=tangent_vec_b,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec_a", "tangent_vec_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_connection(self, tangent_vec_a, tangent_vec_b, base_point, expected, atol):
        res = self.space.metric.connection(tangent_vec_a, tangent_vec_b, base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_curvature_at_identity(
        self, tangent_vec_a, tangent_vec_b, tangent_vec_c, expected, atol
    ):
        res = self.space.metric.curvature_at_identity(
            tangent_vec_a, tangent_vec_b, tangent_vec_c
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_curvature_at_identity_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_c = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.curvature_at_identity(
            tangent_vec_a, tangent_vec_b, tangent_vec_c
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_a=tangent_vec_a,
                    tangent_vec_b=tangent_vec_b,
                    tangent_vec_c=tangent_vec_c,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec_a", "tangent_vec_b", "tangent_vec_c"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_sectional_curvature_at_identity(
        self, tangent_vec_a, tangent_vec_b, expected, atol
    ):
        res = self.space.metric.sectional_curvature_at_identity(
            tangent_vec_a, tangent_vec_b
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_sectional_curvature_at_identity_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.sectional_curvature_at_identity(
            tangent_vec_a, tangent_vec_b
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_a=tangent_vec_a,
                    tangent_vec_b=tangent_vec_b,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec_a", "tangent_vec_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_curvature_derivative_at_identity(
        self, tangent_vec_a, tangent_vec_b, tangent_vec_c, tangent_vec_d, expected, atol
    ):
        res = self.space.metric.curvature_derivative_at_identity(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, tangent_vec_d
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_curvature_derivative_at_identity_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_c = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_d = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.curvature_derivative_at_identity(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, tangent_vec_d
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_a=tangent_vec_a,
                    tangent_vec_b=tangent_vec_b,
                    tangent_vec_c=tangent_vec_c,
                    tangent_vec_d=tangent_vec_d,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=[
                "tangent_vec_a",
                "tangent_vec_b",
                "tangent_vec_c",
                "tangent_vec_d",
            ],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.vec
    def test_exp_at_identity_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.exp(tangent_vec)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=None,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data, test_fnc_name="test_exp")

    @pytest.mark.random
    def test_log_after_exp_at_identity(self, n_points, atol):
        base_point = repeat_point(self.space.identity, n_reps=n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        point = self.space.exp(tangent_vec)
        tangent_vec_ = self.space.log(point, base_point)

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)

    @pytest.mark.random
    def test_exp_after_log_at_identity(self, n_points, atol):
        base_point = repeat_point(self.space.identity, n_reps=n_points)
        end_point = self.data_generator.random_point(n_points)

        tangent_vec = self.space.metric.log(end_point, base_point)
        end_point_ = self.space.metric.exp(tangent_vec)

        self.assertAllClose(end_point_, end_point, atol=atol)


class InvariantMetricVectorTestCase(_InvariantMetricTestCaseMixins):
    def test_left_exp_from_identity(self, tangent_vec, expected, atol):
        res = self.space.metric.left_exp_from_identity(tangent_vec)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_left_exp_from_identity_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.left_exp_from_identity(tangent_vec)

        vec_data = generate_vectorization_data(
            data=[dict(tangent_vec=tangent_vec, expected=expected, atol=atol)],
            arg_names=["tangent_vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_left_log_from_identity(self, point, expected, atol):
        res = self.space.metric.left_log_from_identity(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_left_exp_from_identity_after_left_log_from_identity(self, n_points, atol):
        end_point = self.data_generator.random_point(n_points)

        tangent_vec = self.space.metric.left_log_from_identity(end_point)
        end_point_ = self.space.metric.left_exp_from_identity(tangent_vec)

        self.assertAllClose(end_point_, end_point, atol=atol)

    @pytest.mark.random
    def test_left_log_from_identity_after_left_exp_from_identity(self, n_points, atol):
        base_point = repeat_point(self.space.identity, n_reps=n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        end_point = self.space.metric.left_exp_from_identity(tangent_vec)
        tangent_vec_ = self.space.metric.left_log_from_identity(end_point)

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)

    def test_exp_from_identity(self, tangent_vec, expected, atol):
        res = self.space.metric.exp_from_identity(tangent_vec)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_exp_from_identity_vec(self, n_reps, atol):
        base_point = self.space.identity
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.exp_from_identity(tangent_vec)

        vec_data = generate_vectorization_data(
            data=[dict(tangent_vec=tangent_vec, expected=expected, atol=atol)],
            arg_names=["tangent_vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_log_from_identity(self, point, expected, atol):
        res = self.space.metric.log_from_identity(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_exp_from_identity_after_log_from_identity(self, n_points, atol):
        end_point = self.data_generator.random_point(n_points)

        tangent_vec = self.space.metric.log_from_identity(end_point)
        end_point_ = self.space.metric.exp_from_identity(tangent_vec)

        self.assertAllClose(end_point_, end_point, atol=atol)

    @pytest.mark.random
    def test_log_from_identity_after_exp_from_identity(self, n_points, atol):
        base_point = repeat_point(self.space.identity, n_reps=n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        end_point = self.space.metric.exp_from_identity(tangent_vec)
        tangent_vec_ = self.space.metric.log_from_identity(end_point)

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)


class BiInvariantMetricTestCase(RiemannianMetricTestCase):
    @pytest.mark.random
    def test_invariance(self, n_points, atol):
        group_elem_1 = self.data_generator.random_point(n_points)
        group_elem_2 = self.data_generator.random_point(n_points)

        tangent_vec_11 = self.data_generator.random_tangent_vec(group_elem_1)
        tangent_vec_12 = self.data_generator.random_tangent_vec(group_elem_1)

        dleft_2 = self.space.tangent_translation_map(
            group_elem_2, left=True, inverse=False
        )
        dright_2 = self.space.tangent_translation_map(
            group_elem_2, left=False, inverse=False
        )

        group_elem_3 = self.space.compose(group_elem_2, group_elem_1)
        group_elem_4 = self.space.compose(group_elem_1, group_elem_2)

        tangent_vec_31 = dleft_2(tangent_vec_11)
        tangent_vec_32 = dleft_2(tangent_vec_12)

        tangent_vec_41 = dright_2(tangent_vec_11)
        tangent_vec_42 = dright_2(tangent_vec_12)

        inner_prod = self.space.metric.inner_product(
            tangent_vec_11, tangent_vec_12, group_elem_1
        )
        inner_prod_left = self.space.metric.inner_product(
            tangent_vec_31, tangent_vec_32, group_elem_3
        )
        inner_prod_right = self.space.metric.inner_product(
            tangent_vec_41, tangent_vec_42, group_elem_4
        )
        self.assertAllClose(inner_prod, inner_prod_left, atol=atol)
        self.assertAllClose(inner_prod, inner_prod_right, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.geometry.base import DiffeomorphicManifold
from geomstats.geometry.diffeo import ReversedDiffeo
from geomstats.information_geometry.gamma import (
    GammaDistributions,
    NaturalToStandardDiffeo,
)
from geomstats.test_cases.geometry.base import VectorSpaceOpenSetTestCase
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase
from geomstats.test_cases.information_geometry.base import (
    InformationManifoldMixinTestCase,
)


def scalar_curvature(base_point):
    kappa = base_point[..., 0]
    return (gs.polygamma(1, kappa) + kappa * gs.polygamma(2, kappa)) / (
        2 * (-1 + kappa * gs.polygamma(1, kappa)) ** 2
    )


class StandardGammaDistributions(DiffeomorphicManifold):
    """Class for the manifold of Gamma distributions in standard coordinates."""

    def __init__(self):
        super().__init__(
            diffeo=ReversedDiffeo(NaturalToStandardDiffeo()),
            image_space=GammaDistributions(equip=False),
            dim=2,
            shape=(2,),
            equip=False,
        )


class GammaDistributionsTestCase(
    InformationManifoldMixinTestCase, VectorSpaceOpenSetTestCase
):
    def _check_sample_belongs_to_support(self, sample, atol):
        self.assertTrue(gs.all(sample > 0.0))

    def test_maximum_likelihood_fit(self, data, expected, atol):
        res = self.space.maximum_likelihood_fit(data)
        self.assertAllClose(res, expected, atol=atol)


class GammaMetricTestCase(RiemannianMetricTestCase):
    def test_jacobian_christoffels(self, base_point, expected, atol):
        res = self.space.metric.jacobian_christoffels(base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_scalar_curvature_against_closed_form(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        res = self.space.metric.scalar_curvature(base_point)
        expected = scalar_curvature(base_point)
        self.assertAllClose(res, expected, atol=atol)



"""Core parametrizer classes for Stratified Spaces."""

import pytest

import geomstats.backend as gs
from geomstats.test.random import get_random_times
from geomstats.test.test_case import TestCase
from geomstats.test_cases.geometry.mixins import (
    DistTestCaseMixins,
    GeodesicBVPTestCaseMixins,
)


class RandomDataGenerator:
    def __init__(self, space):
        self.space = space

    def random_point(self, n_points=1):
        return self.space.random_point(n_points)


class PointTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    @pytest.mark.random
    def test_point_is_equal_to_itself(self, n_points, atol):
        initial_point = self.data_generator.random_point(n_points)

        self.assertTrue(gs.all(initial_point.equal(initial_point, atol=atol)))


class PointSetTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    def test_belongs(self, point, expected, atol):
        res = self.space.belongs(point, atol=atol)
        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_random_point_belongs(self, n_points, atol):
        random_point = self.space.random_point(n_points)

        expected = gs.ones(n_points)
        self.test_belongs(random_point, expected, atol=atol)


class PointSetMetricTestCase(DistTestCaseMixins, TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    @pytest.mark.random
    def test_geodesic_boundary_points(self, n_points, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = gs.array([0.0, 1.0])

        geod_func = self.space.metric.geodesic(initial_point, end_point)

        geod_points = geod_func(time)

        if n_points == 1:
            initial_point = [initial_point]
            end_point = [end_point]
            geod_points = [geod_points]

        for geod_point_, initial_point_, end_point_ in zip(
            geod_points, initial_point, end_point
        ):
            self.assertTrue(initial_point_.equal(geod_point_[0], atol=atol))
            self.assertTrue(end_point_.equal(geod_point_[1], atol=atol))

    @pytest.mark.random
    def test_geodesic_bvp_reverse(self, n_points, n_times, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = get_random_times(n_times)

        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)
        geod_func_reverse = self.space.metric.geodesic(
            end_point, end_point=initial_point
        )

        geod_point = geod_func(time)
        geod_point_reverse = geod_func_reverse(1.0 - time)

        if n_points == 1:
            geod_point = [geod_point]
            geod_point_reverse = [geod_point_reverse]

        for geod_point_, geod_point_reverse_ in zip(geod_point, geod_point_reverse):
            self.assertTrue(gs.all(geod_point_.equal(geod_point_reverse_, atol)))


class PointSetMetricWithArrayTestCase(
    DistTestCaseMixins, GeodesicBVPTestCaseMixins, TestCase
):
    tangent_to_multiple = False
    is_metric = True

    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    def test_geodesic(
        self,
        initial_point,
        end_point,
        time,
        expected,
        atol,
    ):
        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)
        res = geod_func(time)
        self.assertAllClose(res, expected, atol=atol)



import functools
from types import MethodType

import numpy as np
from scipy.interpolate import PPoly

import geomstats.backend as gs


def result_to_backend_type(result):
    """Convert np.array to gs.array within result object."""
    if gs.__name__.endswith("numpy") or gs.__name__.endswith("autograd"):
        return result

    for key, value in result.items():
        if type(value) is np.ndarray:
            result[key] = gs.from_numpy(value)

        if isinstance(value, PPoly):
            new_ppoly = _InstanceConvertOutputWrapper(value)
            result[key] = new_ppoly

    return result


def _convert_np_output(func):
    @functools.wraps(func)
    def _wrapped(*args, **kwargs):
        out = func(*args, **kwargs)

        if type(out) is np.ndarray:
            return gs.from_numpy(out)

        return out

    return _wrapped


class _InstanceConvertOutputWrapper:
    """Dynamic wrapper for an instance to convert method output to gs.array."""

    def __init__(self, instance):
        self._instance = instance
        self._dict = {}

    def __getattr__(self, attr_name):
        if attr_name in self._dict:
            return self._dict[attr_name]

        attr = getattr(self._instance, attr_name)
        if isinstance(attr, MethodType):
            attr = _convert_np_output(attr)
            self._dict[attr_name] = attr

        return attr

    def __call__(self, *args, **kwargs):
        out = self._instance(*args, **kwargs)
        if type(out) is np.ndarray:
            return gs.from_numpy(out)

        return out

    def __dir__(self):
        return dir(self._instance)

    def __repr__(self):
        return repr(self._instance)

    def __str__(self):
        return str(self._instance)



"""Pytorch based linear algebra backend."""

import numpy as _np
import scipy as _scipy
import torch as _torch
from torch.linalg import (
    cholesky,
    det,
    eig,
    eigh,
    eigvalsh,
    inv,
    matrix_power,
    qr,
    solve,
)
from torch.linalg import matrix_exp as expm

from .._backend_config import np_atol as atol
from ..numpy import linalg as _gsnplinalg
from ._dtype import _cast_out_to_input_dtype


class _Logm(_torch.autograd.Function):
    """Torch autograd function for matrix logarithm.

    Implementation based on:
    https://github.com/pytorch/pytorch/issues/9983#issuecomment-891777620
    """

    @staticmethod
    def _logm(x):
        np_logm = _gsnplinalg.logm(x.detach().cpu())
        torch_logm = _torch.from_numpy(np_logm).to(x.device, dtype=x.dtype)
        return torch_logm

    @staticmethod
    def forward(ctx, tensor):
        """Apply matrix logarithm to a tensor."""
        ctx.save_for_backward(tensor)
        return _Logm._logm(tensor)

    @staticmethod
    def backward(ctx, grad):
        """Run gradients backward."""
        (tensor,) = ctx.saved_tensors

        vectorized = tensor.ndim == 3
        axes = (0, 2, 1) if vectorized else (1, 0)
        tensor_H = tensor.permute(axes).conj().to(grad.dtype)
        n = tensor.size(-1)
        bshape = tensor.shape[:-2] + (2 * n, 2 * n)
        backward_tensor = _torch.zeros(*bshape, dtype=grad.dtype, device=grad.device)
        backward_tensor[..., :n, :n] = tensor_H
        backward_tensor[..., n:, n:] = tensor_H
        backward_tensor[..., :n, n:] = grad

        return _Logm._logm(backward_tensor).to(tensor.dtype)[..., :n, n:]


logm = _Logm.apply


def sqrtm(x):
    np_sqrtm = _np.vectorize(_scipy.linalg.sqrtm, signature="(n,m)->(n,m)")(x)
    if np_sqrtm.dtype.kind == "c":
        np_sqrtm = np_sqrtm.astype(f"complex{int(np_sqrtm.dtype.name[7:]) // 2}")

    return _torch.from_numpy(np_sqrtm)


def svd(x, full_matrices=True, compute_uv=True):
    if compute_uv:
        return _torch.linalg.svd(x, full_matrices=full_matrices)

    return _torch.linalg.svdvals(x)


def norm(x, ord=None, axis=None):
    if axis is None:
        return _torch.linalg.norm(x, ord=ord)
    return _torch.linalg.norm(x, ord=ord, dim=axis)


def matrix_rank(a, hermitian=False, **_unused_kwargs):
    return _torch.linalg.matrix_rank(a, hermitian=hermitian)


def quadratic_assignment(a, b, options):
    return list(_scipy.optimize.quadratic_assignment(a, b, options=options).col_ind)


def solve_sylvester(a, b, q):
    if (
        a.shape == b.shape
        and _torch.all(a == b)
        and _torch.all(_torch.abs(a - a.transpose(-2, -1)) < 1e-6)
    ):
        eigvals, eigvecs = eigh(a)
        if _torch.all(eigvals >= 1e-6):
            tilde_q = eigvecs.transpose(-2, -1) @ q @ eigvecs
            tilde_x = tilde_q / (eigvals[..., :, None] + eigvals[..., None, :])
            return eigvecs @ tilde_x @ eigvecs.transpose(-2, -1)

        conditions = _torch.all(eigvals >= 1e-6) or (
            a.shape[-1] >= 2.0
            and _torch.all(eigvals[..., 0] > -1e-6)
            and _torch.all(eigvals[..., 1] >= 1e-6)
            and _torch.all(_torch.abs(q + q.transpose(-2, -1)) < 1e-6)
        )
        if conditions:
            tilde_q = eigvecs.transpose(-2, -1) @ q @ eigvecs
            tilde_x = tilde_q / (
                eigvals[..., :, None] + eigvals[..., None, :] + _torch.eye(a.shape[-1])
            )
            return eigvecs @ tilde_x @ eigvecs.transpose(-2, -1)

    solution = _np.vectorize(
        _scipy.linalg.solve_sylvester, signature="(m,m),(n,n),(m,n)->(m,n)"
    )(a, b, q)
    return _torch.from_numpy(solution)


# (TODO) (sait) _torch.linalg.cholesky_ex for even faster way
def is_single_matrix_pd(mat):
    """Check if 2D square matrix is positive definite."""
    if mat.shape[0] != mat.shape[1]:
        return False
    if mat.dtype in [_torch.complex64, _torch.complex128]:
        is_hermitian = _torch.all(
            _torch.abs(mat - _torch.conj(_torch.transpose(mat, 0, 1))) < atol
        )
        if not is_hermitian:
            return False
        eigvals = _torch.linalg.eigvalsh(mat)
        return _torch.min(_torch.real(eigvals)) > 0
    try:
        _torch.linalg.cholesky(mat)
        return True
    except RuntimeError:
        return False


@_cast_out_to_input_dtype
def fractional_matrix_power(A, t):
    """Compute the fractional power of a matrix."""
    if A.ndim == 2:
        out = _scipy.linalg.fractional_matrix_power(A, t)
    else:
        out = _np.stack([_scipy.linalg.fractional_matrix_power(A_, t) for A_ in A])

    return _torch.tensor(out)


def polar(*args, **kwargs):
    """Polar decomposition of a matrix."""
    func = _np.vectorize(
        _scipy.linalg.polar, signature="(n,n)->(n,n),(n,n)", excluded=["side"]
    )
    unitary, hermitian = func = func(*args, **kwargs)

    return _torch.from_numpy(unitary), _torch.from_numpy(hermitian)



"""Statistical Manifold of beta distributions with the Fisher metric.

Lead author: Alice Le Brigant.
"""

from scipy.stats import beta

import geomstats.backend as gs
from geomstats.information_geometry.base import ScipyUnivariateRandomVariable
from geomstats.information_geometry.dirichlet import (
    DirichletDistributions,
    DirichletMetric,
)


class BetaDistributions(DirichletDistributions):
    r"""Class for the manifold of beta distributions.

    This is Beta = :math:`R_+^* \times R_+^*`, the upper-right
    quadrant of the Euclidean plane.

    Attributes
    ----------
    dim : int
        Dimension of the manifold of beta distributions, equal to 2.
    embedding_space : Manifold
        Embedding manifold.
    """

    def __init__(self, equip=True):
        super().__init__(dim=2, equip=equip)
        self.support_shape = ()
        self._scp_rv = BetaDistributionsRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return BetaMetric

    def sample(self, point, n_samples=1):
        """Sample from the beta distribution.

        Sample from the beta distribution with parameters provided by point.
        This gives samples in the segment [0, 1].

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point representing a beta distribution.
        n_samples : int
            Number of points to sample with each pair of parameters in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples]
            Sample from beta distributions.)
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        """Compute pdf associated to point.

        Compute the probability density function of the beta
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point representing a beta distribution.

        Returns
        -------
        pdf : function
            Probability density function of the beta distribution with
            parameters provided by point.
        """
        alpha = gs.expand_dims(point[..., 0], axis=-1)
        beta = gs.expand_dims(point[..., 1], axis=-1)

        def pdf(x):
            """Generate parameterized function for normal pdf.

            Parameters
            ----------
            x : array-like, shape=[n_samples,]
                Points at which to compute the probability density function.

            Returns
            -------
            pdf_at_x : array-like, shape=[..., n_samples]
                Values of pdf at x for each value of the parameters provided
                by point.
            """
            x = gs.reshape(gs.array(x), (-1,))
            return (
                x ** (alpha - 1)
                * (1 - x) ** (beta - 1)
                / (gs.gamma(alpha) * gs.gamma(beta) / gs.gamma(alpha + beta))
            )

        return pdf

    @staticmethod
    def maximum_likelihood_fit(data, loc=0, scale=1, epsilon=1e-6):
        """Estimate parameters from samples.

        This a wrapper around scipy's maximum likelihood estimator to
        estimate the parameters of a beta distribution from samples.

        Parameters
        ----------
        data : array-like, shape=[..., n_samples]
            Data to estimate parameters from. Arrays of
            different length may be passed.
        loc : float
            Location parameter of the distribution to estimate parameters
            from. It is kept fixed during optimization.
            Optional, default: 0.
        scale : float
            Scale parameter of the distribution to estimate parameters
            from. It is kept fixed during optimization.
            Optional, default: 1.

        Returns
        -------
        parameter : array-like, shape=[..., 2]
            Estimate of parameter obtained by maximum likelihood.
        """
        data = gs.where(data == 1.0, 1.0 - epsilon, data)
        data = gs.where(data == 0.0, epsilon, data)
        data = gs.to_ndarray(data, to_ndim=2)
        parameters = []
        for sample in data:
            param_a, param_b, _, _ = beta.fit(sample, floc=loc, fscale=scale)
            parameters.append(gs.array([param_a, param_b]))
        return parameters[0] if len(data) == 1 else gs.stack(parameters)


class BetaMetric(DirichletMetric):
    """Class for the Fisher information metric on beta distributions."""

    @staticmethod
    def metric_det(point):
        """Compute the determinant of the metric.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point representing a beta distribution.


        Returns
        -------
        metric_det : array-like, shape=[...,]
            Determinant of the metric.
        """
        param_a = point[..., 0]
        param_b = point[..., 1]
        metric_det = gs.polygamma(1, param_a) * gs.polygamma(1, param_b) - gs.polygamma(
            1, param_a + param_b
        ) * (gs.polygamma(1, param_a) + gs.polygamma(1, param_b))
        return metric_det


class BetaDistributionsRandomVariable(ScipyUnivariateRandomVariable):
    """A beta random variable."""

    def __init__(self, space):
        super().__init__(space, beta.rvs, beta.pdf)

    @staticmethod
    def _flatten_params(point, pre_flat_shape):
        param_a = gs.expand_dims(point[..., 0], axis=-1)
        param_b = gs.expand_dims(point[..., 1], axis=-1)

        flat_param_a = gs.reshape(gs.broadcast_to(param_a, pre_flat_shape), (-1,))
        flat_param_b = gs.reshape(gs.broadcast_to(param_b, pre_flat_shape), (-1,))
        return {"a": flat_param_a, "b": flat_param_b}



import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.test_cases.geometry.stratified.point_set import PointTestCase


class RandomGroveDataGenerator(RandomDataGenerator):
    def __init__(self, space, topology):
        super().__init__(space)
        self.topology = topology

    def random_point(self, n_points=1):
        return self.space.random_grove_point(self.topology, n_points)


class WaldTestCase(PointTestCase):
    def test_corr(self, point, expected, atol):
        self.assertAllClose(point.corr, expected, atol=atol)


class WaldGeodesicSolverTestCase(TestCase):
    @pytest.mark.random
    def test_discrete_geodesic_reverse(self, n_points, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        geod_points = self.geodesic_solver.discrete_geodesic(initial_point, end_point)
        geod_points_reversed = self.geodesic_solver.discrete_geodesic(
            end_point, initial_point
        )

        if n_points == 1:
            geod_points = [geod_points]
            geod_points_reversed = [geod_points_reversed]

        for geod_points_, geod_points_reversed_ in zip(
            geod_points, geod_points_reversed
        ):
            geod_points_reversed_.reverse()
            self.assertTrue(
                gs.all(geod_points_.equal(geod_points_reversed_, atol=atol))
            )



"""Initial value problem solvers implementation."""

from abc import ABC, abstractmethod

import scipy

import geomstats.backend as gs
import geomstats.integrator as gs_integrator
from geomstats.errors import check_parameter_accepted_values
from geomstats.numerics._common import result_to_backend_type


def _merge_scipy_results(results, same_t=False):
    keys = list(results[0].keys())
    merged_results = {key: [] for key in keys}

    for result in results:
        for key, value in merged_results.items():
            value.append(result[key])

    if same_t:
        merged_results["t"] = merged_results["t"][0]
        merged_results["y"] = gs.stack(merged_results["y"])

    return merged_results


class OdeResult(scipy.optimize.OptimizeResult):
    """Bunch object (follows scipy).

    Its purposes is to homogenize output of different integrators.
    """

    def get_last_y(self):
        """Get value for last y.

        Allows to have y represented as an `gs.array` or `list[gs.array]` (latter
        for cases where they have different shapes).

        Assumes last `t` is the same.
        """
        if isinstance(self.y, (list, tuple)):
            return gs.stack([y_[-1] for y_ in self.y])

        return self.y[-1]


class ODEIVPIntegrator(ABC):
    """Abstract class for ode ivp solvers.

    Parameters
    ----------
    save_result : bool
        If True, result is stored after calling `integrate` or `integrate_t`.
    tchosen : bool
        Informs about ability to solve at chosen times.
        If False, then does not implement `integrate_t`.
    """

    def __init__(self, save_result=False, tchosen=False):
        self.save_result = save_result
        self.tchosen = tchosen

        self.result_ = None

    @abstractmethod
    def integrate(self, force, initial_state, end_time):
        """Integrate force.

        Parameters
        ----------
        force : callable
            Function to integrate: `f(state, t)`.
        initial_state : array-like, shape=[..., n_vars, *point_shape]
            Initial state.
        end_time : float or None
            Integration end time.

        Returns
        -------
        result : OdeResult
        """

    def integrate_t(self, force, initial_state, t_eval):
        """Integrate force while choosing evaluating points.

        Parameters
        ----------
        force : callable
            Function to integrate: `f(state, t)`.
        initial_state : array-like, shape=[..., n_vars, *point_shape]
            Initial state.
        t_eval : array-like
            Times at which to store the computed solution.

        Returns
        -------
        result : OdeResult
        """
        raise NotImplementedError("Can't solve for chosen evaluating points.")


class GSIVPIntegrator(ODEIVPIntegrator):
    """In-house ODE integrator.

    Parameters
    ----------
    n_steps : int
        Number of steps to perform.
    step_type : str
        Type of integration step.
        Possible values are `euler`, `rk2`, `rk4`.
    save_result : bool
        If True, result is stored after calling `integrate` or `integrate_t`.
    """

    def __init__(self, n_steps=10, step_type="euler", save_result=False):
        super().__init__(save_result=save_result, tchosen=False)
        self.step_type = step_type
        self.n_steps = n_steps

    @property
    def step_type(self):
        """Integrator step type."""
        return self._step_type

    @step_type.setter
    def step_type(self, value):
        if callable(value):
            step_function = value
            value = None
        else:
            check_parameter_accepted_values(
                value, "step_type", gs_integrator.STEP_FUNCTIONS
            )
            step_function = getattr(gs_integrator, gs_integrator.STEP_FUNCTIONS[value])

        self._step_function = step_function
        self._step_type = value

    def _step(self, force, state, time, dt):
        return self._step_function(force, state, time, dt)

    def _get_n_fevals(self, n_steps):
        n_evals_step = gs_integrator.FEVALS_PER_STEP[self.step_type]
        return n_evals_step * n_steps

    def _integrate(self, force, initial_state, end_time=1.0):
        dt = end_time / self.n_steps
        states = [initial_state]
        current_state = initial_state

        for i in range(self.n_steps):
            current_state = self._step(
                force=force, state=current_state, time=i * dt, dt=dt
            )
            states.append(current_state)

        return states

    def integrate(self, force, initial_state, end_time=1.0):
        """Integrate force.

        Parameters
        ----------
        force : callable
            Function to integrate: `f(state, t)`.
        initial_state : array-like, shape=[..., n_vars, *point_shape]
            Initial state.
        end_time : float or None
            Integration end time.

        Returns
        -------
        result : OdeResult
        """
        states = self._integrate(force, initial_state, end_time=end_time)

        ts = gs.linspace(0.0, end_time, self.n_steps + 1)
        nfev = self._get_n_fevals(self.n_steps)

        result = OdeResult(t=ts, y=gs.array(states), nfev=nfev, njev=0, sucess=True)

        if self.save_result:
            self.result_ = result

        return result


class ScipySolveIVP(ODEIVPIntegrator):
    """Wrapper for scipy.integrate.solve_ivp.

    Check
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html
    for additional options.

    Parameters
    ----------
    method : str
        Integration method.
    save_result : bool
        If True, result is stored after calling `integrate` or `integrate_t`.
    point_ndim = int
        Dimension of array representing a point in the space.
    """

    def __init__(self, method="RK45", save_result=False, point_ndim=1, **options):
        super().__init__(save_result=save_result, tchosen=True)
        self.method = method
        self.point_ndim = point_ndim
        self.options = options

    def _integrate(self, force, initial_state, end_time=1.0, t_eval=None):
        if initial_state.ndim > (self.point_ndim + 1):
            results = []
            for initial_state_ in initial_state:
                results.append(
                    self._integrate_single(force, initial_state_, end_time, t_eval)
                )

            result = OdeResult(_merge_scipy_results(results, same_t=t_eval is not None))

        else:
            result = self._integrate_single(
                force, initial_state, end_time, t_eval=t_eval
            )
            result = OdeResult(**result)

        if self.save_result:
            self.result_ = result

        return result

    def integrate(self, force, initial_state, end_time=1.0):
        """Integrate force.

        Parameters
        ----------
        force : callable
            Function to integrate: `f(state, t)`.
        initial_state : array-like, shape=[..., n_vars, *point_shape]
            Initial state.
        end_time : float or None
            Integration end time.

        Returns
        -------
        result : OdeResult
        """
        return self._integrate(force, initial_state, end_time=end_time)

    def integrate_t(self, force, initial_state, t_eval):
        """Integrate force at `t_eval` points.

        Parameters
        ----------
        force : callable
            Function to integrate: `f(state, t)`.
        initial_state : array-like, shape=[..., n_vars, *point_shape]
            Initial state.
        t_eval : array-like
            Times at which to store the computed solution.

        Returns
        -------
        result : OdeResult
        """
        return self._integrate(force, initial_state, end_time=t_eval[-1], t_eval=t_eval)

    def _integrate_single(self, force, initial_state, end_time=1.0, t_eval=None):
        def force_(t, state):
            state = gs.from_numpy(state)
            unraveled_state = gs.reshape(state, initial_state.shape)

            return gs.reshape(
                force(unraveled_state, t),
                state.shape,
            )

        raveled_initial_state = gs.reshape(initial_state, (-1,))

        result = scipy.integrate.solve_ivp(
            force_,
            (0.0, end_time),
            raveled_initial_state,
            method=self.method,
            t_eval=t_eval,
            **self.options,
        )
        result = result_to_backend_type(result)
        result.y = gs.reshape(
            gs.moveaxis(result.y, 0, -1),
            (-1,) + initial_state.shape,
        )

        return result



"""Statistical Manifold of Binomial distributions with the Fisher metric.

Lead authors: Jules Deschamps, Tra My Nguyen.
"""

from scipy.stats import expon

import geomstats.backend as gs
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.information_geometry.base import (
    InformationManifoldMixin,
    ScipyUnivariateRandomVariable,
)


class ExponentialDistributions(InformationManifoldMixin, VectorSpaceOpenSet):
    """Class for the manifold of exponential distributions.

    This is the parameter space of exponential distributions
    i.e. the half-line of positive reals.
    """

    def __init__(self, equip=True):
        super().__init__(
            dim=1,
            embedding_space=Euclidean(dim=1, equip=False),
            support_shape=(),
            equip=equip,
        )
        self._scp_rv = ExponentialDistributionsRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return ExponentialMetric

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold of exponential distributions.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point to be checked.
        atol : float
            Tolerance to evaluate positivity.
            Optional, default: gs.atol

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean indicating whether point represents an exponential
            distribution.
        """
        belongs_shape = self.shape == point.shape[-self.point_ndim :]
        if not belongs_shape:
            shape = point.shape[: -self.point_ndim]
            return gs.zeros(shape, dtype=bool)
        return gs.squeeze(point >= -atol)

    def random_point(self, n_samples=1, lower_bound=0.1, upper_bound=1.0):
        """Sample parameters of exponential distributions.

        The uniform distribution on [0, bound] is used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Right-end ot the segment where exponential parameters are sampled.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[n_samples,]
            Sample of points representing exponential distributions.
        """
        size = (n_samples, self.dim) if n_samples != 1 else (self.dim,)
        return (upper_bound - lower_bound) * gs.random.rand(*size) + lower_bound

    def projection(self, point, atol=gs.atol):
        """Project a point in ambient space to the open set.

        The last coordinate is floored to `gs.atol` if it is non-positive.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point in ambient space.
        atol : float
            Tolerance to evaluate positivity.

        Returns
        -------
        projected : array-like, shape=[..., 1]
            Projected point.
        """
        return gs.where(point < atol, atol, point)

    def sample(self, point, n_samples=1):
        """Sample from the exponential distribution.

        Sample from the exponential distribution with parameter provided
        by point.

        Parameters
        ----------
        point : array-like, shape=[...,]
            Point representing an exponential distribution.
        n_samples : int
            Number of points to sample with each parameter in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples]
            Sample from exponential distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        """Compute pdf associated to point.

        Compute the probability density function of the exponential
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point representing an exponential distribution (scale).

        Returns
        -------
        pdf : function
            Probability density function of the exponential distribution with
            scale parameter provided by point.
        """

        def pdf(x):
            """Generate parameterized function for exponential pdf.

            The pdf is parameterized by the scale parameter of the exponential,
            which is equal to 1 / lambda, where lambda is the rate parameter.

            Parameters
            ----------
            x : array-like, shape=[n_points,]
                Points at which to compute the probability density function.

            Returns
            -------
            pdf_at_x : array-like, shape=[..., n_points]
            """
            x = gs.reshape(gs.array(x), (-1,))
            pdf_at_x = point * gs.exp(-point * x)
            return gs.where(x >= 0, pdf_at_x, 0.0)

        return pdf


class ExponentialMetric(RiemannianMetric):
    """Class for the Fisher information metric on exponential distributions.

    References
    ----------
    .. [AM1981] Atkinson, C., & Mitchell, A. F. (1981). Rao's distance measure.
        Sankhyā: The Indian Journal of Statistics, Series A, 345-365.
    """

    def squared_dist(self, point_a, point_b):
        """Compute squared distance associated with the exponential Fisher Rao metric.

        Parameters
        ----------
        point_a : array-like, shape=[..., 1]
            Point representing an exponential distribution (scale parameter).
        point_b : array-like, shape=[..., 1]
            Point representing a exponential distribution (scale parameter).

        Returns
        -------
        squared_dist : array-like, shape=[...,]
            Squared distance between points point_a and point_b.
        """
        return gs.squeeze(gs.log(point_a / point_b) ** 2)

    def metric_matrix(self, base_point):
        """Compute the metric matrix at the tangent space at base_point.

        Parameters
        ----------
        base_point : array-like, shape=[..., 1]
            Point representing a binomial distribution.

        Returns
        -------
        mat : array-like, shape=[..., 1, 1]
            Metric matrix.
        """
        return gs.expand_dims(1 / base_point**2, axis=-1)

    @staticmethod
    def _geodesic_path(t, initial_point, base):
        """Generate parameterized function for geodesic curve.

        Parameters
        ----------
        t : array-like, shape=[n_times,]
            Times at which to compute points of the geodesics.

        Returns
        -------
        geodesic : array-like, shape=[..., n_times, 1]
            Values of the geodesic at times t.
        """
        t = gs.reshape(gs.array(t), (-1,))
        base_aux, t_aux = gs.broadcast_arrays(base, t)
        return gs.expand_dims(initial_point * base_aux**t_aux, axis=-1)

    def _geodesic_ivp(self, initial_point, initial_tangent_vec):
        """Solve geodesic initial value problem.

        Compute the parameterized function for the geodesic starting at
        initial_point with initial velocity given by initial_tangent_vec.

        Parameters
        ----------
        initial_point : array-like, shape=[..., 1]
            Initial point.

        initial_tangent_vec : array-like, shape=[..., 1]
            Tangent vector at initial point.

        Returns
        -------
        path : function
            Parameterized function for the geodesic curve starting at
            initial_point with velocity initial_tangent_vec.
        """
        base = gs.exp(initial_tangent_vec / initial_point)
        return lambda t: self._geodesic_path(t, initial_point, base)

    def _geodesic_bvp(self, initial_point, end_point):
        """Solve geodesic boundary problem.

        Compute the parameterized function for the geodesic starting at
        initial_point and ending at end_point.

        Parameters
        ----------
        initial_point : array-like, shape=[..., 1]
            Initial point.
        end_point : array-like, shape=[..., 1]
            End point.

        Returns
        -------
        path : function
            Parameterized function for the geodesic curve starting at
            initial_point and ending at end_point.
        """
        base = end_point / initial_point
        return lambda t: self._geodesic_path(t, initial_point, base)

    def exp(self, tangent_vec, base_point):
        """Compute exp map of a base point in tangent vector direction.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 1]
            Tangent vector at base point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., 1]
            End point of the geodesic starting at base_point with
            initial velocity tangent_vec.
        """
        return gs.exp(tangent_vec / base_point) * base_point

    def log(self, point, base_point):
        """Compute log map using a base point and an end point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            End point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., 1]
            Initial velocity of the geodesic starting at base_point and
            reaching point at time 1.
        """
        return base_point * gs.log(point / base_point)


class ExponentialDistributionsRandomVariable(ScipyUnivariateRandomVariable):
    """An exponential random variable."""

    def __init__(self, space):
        super().__init__(space, expon.rvs, expon.pdf)

    @staticmethod
    def _flatten_params(point, pre_flat_shape):
        flat_scale_param = gs.reshape(gs.broadcast_to(1 / point, pre_flat_shape), (-1,))
        return {"scale": flat_scale_param}



"""Align All and Compute for Graph Space.

Lead author: Anna Calissano.
"""

import logging
import random

from sklearn.base import BaseEstimator

import geomstats.backend as gs
from geomstats.errors import check_parameter_accepted_values
from geomstats.learning._sklearn import PCA, LinearRegression
from geomstats.learning.frechet_mean import FrechetMean


def _warn_max_iterations(iteration, max_iter):
    if iteration + 1 == max_iter:
        logging.warning(
            f"Maximum number of iterations {max_iter} reached. "
            "The estimate may be inaccurate"
        )


class _AACFrechetMean(BaseEstimator):
    r"""Class AAC for Frechet Mean on Graph Space.

    The Align All and Compute (AAC) algorithm for Frechet Mean (FM)estimation is
    introduced in [Calissano2020]_ and it estimates the Frechet Mean for a set of
    labeled or unlabeled graphs. The idea is to optimally aligned the graphs to the
    current mean estimator using the optimal alignment between the graphs and the mean
    (graph matching between two graphs) and compute the mean estimation between the
    aligned adjacency matrices (the arithmetic mean in the euclidean space of dimension
    :math:`nodes \times nodes`). The algorithm stops as soon as the distance between two
    consecutive estimations is lower then :math:`\epsilon` or the maximum number of
    iterations is reached. The initialization step consists in aligning the data
    with respect to an initial point.

    Parameters
    ----------
    space : GraphSpace
        Graph space total space with a quotient structure.
    epsilon: float, default=1e-6
        Stopping criterion for the estimation step, i.e., the distance between two
        consecutive estimators.
    max_iter: int, default = 20
        Stopping criterion on the maximum number of iterations.
    init_point: array-like, shape=[n_nodes, n_nodes] or GraphPoint, default random.
        Algorithm initialization.
    save_last_X: bool, default = True
        Flag to save the data as aligned in the last algorithm iteration.
    total_space_estimator_kwargs : dict
        Total space estimator keyword arguments.

    Attributes
    ----------
    total_space_estimator : BaseEstimator
        Frechet mean estimator in total space.
    estimate_ : array-like, mean=[n_nodes, n_nodes]
        Mean.
    n_iter_ : int
        Number of performed iterations.
    aligned_X_: array-like, shape=[n_samples, n_nodes, n_nodes] or set of GraphPoint.
        Set of aligned data as after the last call of fit.
        Saved if ``save_last_X is True``.

    References
    ----------
    .. [Calissano2020]  Calissano, A., Feragen, A., Vantini, S.
        “Graph Space: Geodesic Principal Components for a Population of
        Network-valued Data.” Mox report 14, 2020.
        https://mox.polimi.it/reports-and-theses/publication-results/?id=855.
    """

    def __init__(
        self,
        space,
        *,
        epsilon=1e-3,
        max_iter=20,
        init_point=None,
        total_space_estimator_kwargs=None,
        save_last_X=True,
    ):
        self.space = space
        self.epsilon = epsilon
        self.max_iter = max_iter
        self.init_point = init_point
        self.save_last_X = save_last_X

        self.total_space_estimator_kwargs = total_space_estimator_kwargs or {}
        self.total_space_estimator = FrechetMean(
            self.space, **self.total_space_estimator_kwargs
        )

        self.estimate_ = None
        self.n_iter_ = None
        self.aligned_X_ = None

    def fit(self, X, y=None):
        """Fit the Frechet Mean.

        Parameters
        ----------
        X : array-like, shape=[n_samples, n_nodes, n_nodes].
            Dataset to estimate the FM.
        y : Ignored
            Ignored.

        Returns
        -------
        self : object
            Returns self.
        """
        previous_estimate = (
            random.choice(X) if self.init_point is None else self.init_point
        )
        aligned_X = X

        for iteration in range(self.max_iter):
            aligned_X = self.space.aligner.align(aligned_X, previous_estimate)
            new_estimate = self.total_space_estimator.fit(aligned_X).estimate_

            error = self.space.metric.dist(previous_estimate, new_estimate)
            if error < self.epsilon:
                break

            previous_estimate = new_estimate
        else:
            _warn_max_iterations(iteration, self.max_iter)

        if self.save_last_X:
            self.aligned_X_ = aligned_X
        self.estimate_ = new_estimate
        self.n_iter_ = iteration

        return self


class _AACGGPCA(BaseEstimator):
    r"""Class AAC for Generalized Geodesic Principal Components (GGPCA) on Graph Space.

    The Align All and Compute (AAC) algorithm for GGPCA estimation is
    introduced in [Calissano2020]_ and it estimates the GGPCA for a set of
    labeled or unlabeled graphs. The idea is to optimally aligned the graphs to the
    current GGPCA estimator using the optimal alignment between the graphs and the
    geodesics and then compute the GGPCA estimation between the aligned adjacency
    matrices (the PCA in the euclidean space of dimension :math:`nodes \times nodes`).
    The algorithm stops as soon as the percentage of variance explained by PCA in two
    consecutive estimations is lower then :math:`\epsilon` or the maximum number of
    iteration is reached. The initialization step consists in aligning all the data with
    respect to an initial point.

    Parameters
    ----------
    space : GraphSpace
        Graph space total space with a quotient structure.
    epsilon: float, default=1e-6
        Stopping criterion for the estimation step, i.e., the distance between two
        consecutive estimators.
    max_iter: int, default = 20
        Stopping criterion on the maximum number of iterations.
    init_point: array-like, shape=[n_nodes, n_nodes] or GraphPoint, default random.
        Algorithm initialization.
    n_components: int
        Number of principal components to be estimated. Notice that the convergence is
        ensured only for the first principal component.
    save_last_X: bool, default = True
        Flag to save the data as aligned in the last algorithm iteration.

    Attributes
    ----------
    total_space_estimator: BaseEstimator
        Method for the estimation of the PCA for a set of flattened adjacency matrices
        in the total space. Check geomstats.learning._sklearn_wrapper for details.
        Default: ``sklearn.decomposition.PCA``.
    n_iter_ : int
        Number of performed iterations.
    aligned_X_: array-like, shape=[n_samples, n_nodes, n_nodes] or set of GraphPoint.
        Set of aligned data as after the last call of fit.
        Saved if ``save_last_X is True``.

    References
    ----------
    .. [Calissano2020]  Calissano, A., Feragen, A., Vantini, S.
        “Graph Space: Geodesic Principal Components for a Population of
        Network-valued Data.” Mox report 14, 2020.
        https://mox.polimi.it/reports-and-theses/publication-results/?id=855.
    """

    def __init__(
        self,
        space,
        *,
        n_components=2,
        epsilon=1e-3,
        max_iter=20,
        init_point=None,
        save_last_X=True,
    ):
        self.space = space
        self.epsilon = epsilon
        self.max_iter = max_iter
        self.init_point = init_point
        self.n_components = n_components
        self.save_last_X = save_last_X

        self.total_space_estimator = PCA(n_components=self.n_components)
        self.n_iter_ = None
        self.aligned_X_ = None

    @property
    def components_(self):
        """Principal Components in the total space.

        GGPCA expressed as vectors in the total space.
        """
        return self.total_space_estimator.reshaped_components_

    @property
    def explained_variance_(self):
        """Variance Explained along the GGPCA."""
        return self.total_space_estimator.explained_variance_

    @property
    def explained_variance_ratio_(self):
        """Percentage of Variance Explained along the GGPCA."""
        return self.total_space_estimator.explained_variance_ratio_

    @property
    def mean_(self):
        """Mean at the last iteration."""
        return self.total_space_estimator.reshaped_mean_

    def fit(self, X, y=None):
        """Fit the GGPCA.

        Parameters
        ----------
        X : array-like, shape=[n_samples, n_nodes, n_nodes] or set of GraphPoint.
            Dataset to estimate the GGPCA.
        y : Ignored
            Ignored.

        Returns
        -------
        self : object
            Returns self.

        Note: Default method in the total space is sklearn.decomposition.PCA where
        the input data are centered but not scaled for each feature.
        """
        x = random.choice(X) if self.init_point is None else self.init_point
        aligned_X = self.space.aligner.align(X, x)

        self.total_space_estimator.fit(aligned_X)
        previous_expl = self.total_space_estimator.explained_variance_ratio_[0]

        for iteration in range(self.max_iter):
            mean = self.total_space_estimator.reshaped_mean_
            direc = self.total_space_estimator.reshaped_components_[0]

            geodesic = self.space.metric.geodesic(
                initial_point=mean, initial_tangent_vec=direc
            )

            aligned_X = self.space.aligner.align_point_to_geodesic(geodesic, aligned_X)
            self.total_space_estimator.fit(aligned_X)
            expl_ = self.total_space_estimator.explained_variance_ratio_[0]

            error = gs.abs(expl_ - previous_expl)
            if error < self.epsilon:
                break
            previous_expl = expl_
        else:
            _warn_max_iterations(iteration, self.max_iter)

        if self.save_last_X:
            self.aligned_X_ = aligned_X
        self.n_iter_ = iteration

        return self


class _AACRegression(BaseEstimator):
    r"""Class AAC for Generalized Geodesic Regression (GGR) on Graph Space.

    The Align All and Compute (AAC) algorithm for GGR estimation is
    introduced in [Calissano2022] and it estimates the GGR for
    :math:`\{(s_i, X_i)\in \mathbb{R}^p\times X/T\}` a set of labeled or unlabeled
    graphs as output and a set of scalar or vector as input:
    :math:`f: \mathbb{R}^p \rightarrow X/T`. The idea is to iteratively estimate a OLS
    regression model between a set of regressors and a set of flattened adjacency
    matrices and align the graphs to the current GGR estimator using the optimal
    alignment for regression. The optimal alignment for regression consists in aligning
    the graph with the corresponding predicted graph along the regression model to
    decrease the prediction error. The algorithm stops as soon as the loss in two
    consecutive estimations is lower then math:`\epsilon` or the maximum number of
    iteration is reached. The initialization step consists in aligning all the data with
    respect to a initial point.

    Parameters
    ----------
    space : GraphSpace
        Graph space total space with a quotient structure.
    epsilon: float, default=1e-6
        Stopping criterion for the estimation step, i.e., the distance between loss
        function in two consecutive estimation steps.
    max_iter: int, default = 20
        Stopping criterion on the maximum number of iterations.
    init_point: array-like, shape=[n_nodes, n_nodes] or GraphPoint, default random.
        Algorithm initialization.
    save_last_y: bool, default=True
        Flag to save the data as aligned in the last algorithm iteration.
    total_space_estimator_kwargs : dict
        Total space estimator keyword arguments.

    Attributes
    ----------
    total_space_estimator: BaseEstimator
        Method for the estimation of the OLS Regression for a set of flattened adjacency
        matrices in the total space.
        Check geomstats.learning._sklearn_wrapper for details.
        Default: ``sklearn.linear_model.LinearRegression``.
    aligned_y_: array-like, shape=[n_samples, n_nodes, n_nodes] or set of GraphPoint.
        Set of aligned data as after the last call of fit.
        Saved if ``self.save_last_y is True``.

    References
    ----------
    .. [Calissano2022]  Calissano, A., Feragen, A., Vantini, S.
        “Graph-valued regression: prediction of unlabelled networks in a non-Euclidean
        Graph Space.”Journal of Multivariate Analysis 190 - 104950, (2022).
        https://doi.org/10.1016/j.jmva.2022.104950.
    """

    def __init__(
        self,
        space,
        *,
        epsilon=1e-3,
        max_iter=20,
        init_point=None,
        total_space_estimator_kwargs=None,
        save_last_y=True,
    ):
        self.space = space
        self.epsilon = epsilon
        self.max_iter = max_iter
        self.init_point = init_point
        self.save_last_y = save_last_y

        self.total_space_estimator_kwargs = total_space_estimator_kwargs or {}
        self.total_space_estimator = LinearRegression(
            **self.total_space_estimator_kwargs
        )
        self.n_iter_ = None
        self.aligned_y_ = None

    def fit(self, X, y):
        """Fit the Generalized Geodesic Regression.

        Parameters
        ----------
        X : array-like, shape=[n_samples, p].
            Dataset of regressors to estimate the GGR.
        y : array-like, shape=[n_samples, n_nodes, n_nodes] or set of GraphPoint.
            Dataset to estimate the GGR.

        Returns
        -------
        self : object
            Returns self.
        """
        y_ = random.choice(y) if self.init_point is None else self.init_point
        aligned_y = self.space.aligner.align(y, y_)

        previous_pred_dist = 1e6
        for iteration in range(self.max_iter):
            self.total_space_estimator.fit(X, aligned_y)
            y_pred = self.total_space_estimator.predict(X)

            aligned_y = self.space.aligner.align(aligned_y, y_pred)
            pred_dist = gs.sum(self.space.metric.dist(y_pred, aligned_y))

            dist_diff = gs.abs(previous_pred_dist - pred_dist)
            if dist_diff < self.epsilon:
                break

            previous_pred_dist = pred_dist
        else:
            _warn_max_iterations(iteration, self.max_iter)

        if self.save_last_y:
            self.aligned_y_ = aligned_y
        self.n_iter_ = iteration

        return self

    def predict(self, X):
        """Predict using the generalized geodesic regression.

        Predict a graph or a set of graphs corresponding to the given regressors. It
        uses the total space prediction.

        Parameters
        ----------
        X : array-like, shape=[n_samples, n_nodes, n_nodes] or set of GraphPoint
            Dataset to estimate the GGR.

        Returns
        -------
        prediction : array-like, shape=[n_samples, n_nodes, n_nodes] or set of
            GraphPoint
            Predicted unlabeled graphs.
        """
        return self.total_space_estimator.predict(X)


class AAC:
    r"""Class for Align all and Compute algorithm on Graph Space.

    The Align All and Compute (AAC) algorithm is introduced in [Calissano2020]_ and it
    allows to compute different statistical estimators: the Frechet Mean, the
    Generalized Geodesic Principal components and the Regression for a set of labeled or
    unlabeled graphs.
    The idea is to optimally aligned the graphs to the current
    estimator using the correct alignment technique and compute the current estimation
    using the geometrical property of the total space, i.e., the Euclidean space of
    adjacency matrices.

    Parameters
    ----------
    space : GraphSpace
        Graph space total space with a quotient structure.
    estimate : str
        Desired estimator. One of the following:

        - "frechet_mean": Frechet Mean estimation [Calissano2020]_
        - "ggpca": Generalized Geodesic Principal Components [Calissano2020]_
        - "regression": Graph-on-vector regression model [Calissano2022]_

    Examples
    --------
    Available example on Graph Space:
    :mod:`notebooks.19_practical_methods__aac`

    Available example on Graph Space with real world data:
    :mod:`notebooks.20_real_world_application__graph_space`

    References
    ----------
    .. [Calissano2020]  Calissano, A., Feragen, A., Vantini, S.
        “Graph Space: Geodesic Principal Components for a Population of
        Network-valued Data.” Mox report 14, 2020.
        https://mox.polimi.it/reports-and-theses/publication-results/?id=855.
    .. [Calissano2022]  Calissano, A., Feragen, A., Vantini, S.
        “Graph-valued regression: prediction of unlabelled networks in a non-Euclidean
        Graph Space.”Journal of Multivariate Analysis 190 - 104950, (2022).
        https://doi.org/10.1016/j.jmva.2022.104950.
    """

    MAP_ESTIMATE = {
        "frechet_mean": _AACFrechetMean,
        "ggpca": _AACGGPCA,
        "regression": _AACRegression,
    }

    def __new__(cls, space, *args, estimate="frechet", **kwargs):
        """Class for Align all and Compute algorithm on Graph Space."""
        check_parameter_accepted_values(
            estimate, "estimate", list(cls.MAP_ESTIMATE.keys())
        )

        return cls.MAP_ESTIMATE[estimate](space, *args, **kwargs)



"""The complex Poincaré disk manifold.

The Poincaré complex disk is a representation of the Hyperbolic space of dimension 2.

Lead author: Yann Cabanes.

References
----------
.. [Cabanes2022] Yann Cabanes. Multidimensional complex stationary
    centered Gaussian autoregressive time series machine learning
    in Poincaré and Siegel disks: application for audio and radar
    clutter classification, PhD thesis, 2022
.. [JV2016] B. Jeuris and R. Vandebril. The Kahler mean of Block-Toeplitz
    matrices with Toeplitz structured blocks, 2016.
    https://epubs.siam.org/doi/pdf/10.1137/15M102112X
.. [Yang2013] Marc Arnaudon, Frédéric Barbaresco and Le Yang.
    Riemannian medians and means with applications to radar signal processing,
    IEEE Journal of Selected Topics in Signal Processing, vol. 7, no. 4, pp. 595-604,
    Aug. 2013, doi: 10.1109/JSTSP.2013.2261798.
    https://ieeexplore.ieee.org/document/6514112
"""

import geomstats.backend as gs
from geomstats.geometry.base import ComplexVectorSpaceOpenSet
from geomstats.geometry.complex_riemannian_metric import ComplexRiemannianMetric
from geomstats.geometry.hermitian import Hermitian


class ComplexPoincareDisk(ComplexVectorSpaceOpenSet):
    """Class for the complex Poincaré disk.

    The Poincaré disk is a representation of the Hyperbolic
    space of dimension 2. Its complex dimension is 1.
    """

    def __init__(self, equip=True):
        super().__init__(
            dim=1,
            embedding_space=Hermitian(dim=1),
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return ComplexPoincareDiskMetric

    @staticmethod
    def belongs(point, atol=gs.atol):
        """Check if a point belongs to the complex unit disk.

        Check if a point belongs to the Poincaré complex disk,
        i.e. Check if its norm is lower than one.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point to be checked.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...]
            Boolean denoting if point belongs to the
            complex Poincaré disk.
        """
        if point.shape[-1] != 1:
            return gs.zeros(point.shape[:-1], dtype=bool)

        return gs.all(gs.abs(point) < 1, axis=-1)

    @staticmethod
    def projection(point):
        """Project a point on the complex unit disk.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point to project.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        projected : array-like, shape=[..., 1]
            Projected point.
        """
        scalars = gs.where(
            gs.abs(point) >= 1 - gs.atol,
            (1 - gs.atol) / gs.abs(point),
            1.0,
        )
        return scalars * point

    @staticmethod
    def random_point(n_samples=1, bound=0.8):
        """Generate random points in the complex unit disk.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample in the tangent space.
            Optional, default: 0.8.

        Returns
        -------
        samples : array-like, shape=[..., 1]
            Points sampled in the unit disk.
        """
        size = (n_samples, 1) if n_samples != 1 else (1,)
        modulus = gs.random.rand(*size, dtype=gs.get_default_cdtype())
        angle = 2 * gs.pi * gs.random.rand(*size, dtype=gs.get_default_cdtype())
        return modulus * gs.exp(1j * angle)


class ComplexPoincareDiskMetric(ComplexRiemannianMetric):
    """Class for the complex Poincaré metric."""

    def metric_matrix(self, base_point):
        """Compute the metric matrix at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        inner_prod_mat : array-like, shape=[...]
            Inner product matrix.
        """
        inner_prod_mat = 1 / (1 - gs.abs(base_point) ** 2) ** 2
        return gs.expand_dims(inner_prod_mat, axis=-1)

    @staticmethod
    def exp(tangent_vec, base_point):
        """Compute the complex Poincaré disk exponential map.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 1]
            Tangent vector at base point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., 1]
            Riemannian exponential.
        """
        base_point, tangent_vec = gs.broadcast_arrays(base_point, tangent_vec)

        theta = gs.angle(tangent_vec)
        s = 2 * gs.abs(tangent_vec) / (1 - gs.abs(base_point) ** 2)

        exp_i_theta = gs.exp(1j * theta)
        exp_minus_s = gs.exp(-s)

        num = base_point + exp_i_theta + (base_point - exp_i_theta) * exp_minus_s
        den = (
            1
            + gs.conj(base_point) * exp_i_theta
            + (1 - gs.conj(base_point) * exp_i_theta) * exp_minus_s
        )
        return num / den

    @staticmethod
    def _tau(point_a, point_b, atol=gs.atol):
        """Compute the coefficient tau.

        The coefficient tau is used to compute the distance
        between point_a and point_b; it is also used to
        compute the logarithm map between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., 1]
            Point.
        point_b : array-like, shape=[..., 1]
            Point.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        tau : array-like, shape=[...]
            Coefficient tau.
        """
        num = gs.abs(point_a - point_b)
        den = gs.maximum(gs.abs(1 - point_a * gs.conj(point_b)), atol)
        delta = gs.squeeze(gs.minimum(num / den, 1 - atol), axis=-1)
        return (1 / 2) * gs.log((1 + delta) / (1 - delta))

    def log(self, point, base_point):
        """Compute the complex Poincaré disk logarithm map.

        Compute the Riemannian logarithm at point base_point,
        of point wrt the positive reals metric.
        This gives a tangent vector at point base_point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        log : array-like, shape=[..., 1]
            Riemannian logarithm.
        """
        angle = gs.angle(point - base_point) - gs.angle(1 - gs.conj(base_point) * point)
        return gs.exp(1j * angle) * gs.einsum(
            "...,...i->...i",
            self._tau(base_point, point),
            (1 - gs.abs(base_point) ** 2),
        )

    def squared_dist(self, point_a, point_b, atol=gs.atol):
        """Compute the complex Poincaré disk squared distance.

        Compute the Riemannian squared distance between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., 1]
            Point.
        point_b : array-like, shape=[..., 1]
            Point.

        Returns
        -------
        squared_dist : array-like, shape=[...]
            Riemannian squared distance.
        """
        dist = self._tau(point_a, point_b, atol=atol)
        return gs.power(dist, 2)

    def dist(self, point_a, point_b, atol=gs.atol):
        """Compute the complex Poincaré disk distance.

        Compute the Riemannian distance between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., 1]
            Point.
        point_b : array-like, shape=[..., 1]
            Point.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        dist : array-like, shape=[...]
            Riemannian distance.
        """
        return self._tau(point_a, point_b, atol=atol)



"""Module providing the SkewSymmetricMatrices class.

This is the Lie algebra of the Special Orthogonal Group.
As basis we choose the matrices with a single 1 on the upper triangular part
of the matrices (and a -1 in its lower triangular part), except in dim 2 and
3 to match usual conventions.

Lead author: Nicolas Guigui.
"""

import geomstats.backend as gs
from geomstats.geometry.lie_algebra import MatrixLieAlgebra
from geomstats.geometry.matrices import Matrices, MatricesMetric


class SkewSymmetricMatrices(MatrixLieAlgebra):
    """Class for skew-symmetric matrices.

    Parameters
    ----------
    n : int
        Number of rows and columns.
    """

    def __init__(self, n, equip=True):
        self.n = n
        dim = int(n * (n - 1) / 2)
        super().__init__(dim=dim, representation_dim=n, equip=equip)
        self.embedding_space = Matrices(n, n, equip=False)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def _create_basis(self):
        """Create the canonical basis."""
        n = self.n
        if n == 2:
            return gs.array([[[0.0, -1.0], [1.0, 0.0]]])
        if n == 3:
            return gs.array(
                [
                    [[0.0, 0.0, 0.0], [0.0, 0.0, -1.0], [0.0, 1.0, 0.0]],
                    [[0.0, 0.0, 1.0], [0.0, 0.0, 0.0], [-1.0, 0.0, 0.0]],
                    [[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                ]
            )
        indices, data = [], []
        k = -1
        for row in range(n - 1):
            for col in range(row + 1, n):
                k += 1
                indices.extend([(k, row, col), (k, col, row)])
                data.extend([1.0, -1.0])

        return gs.array_from_sparse(indices, data, (k + 1, n, n))

    def belongs(self, point, atol=gs.atol):
        """Evaluate if point is a skew-symmetric matrix.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Square matrix to check.
        atol : float
            Tolerance for the equality evaluation.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if matrix is skew symmetric.
        """
        has_right_shape = self.embedding_space.belongs(point)
        if gs.all(has_right_shape):
            return Matrices.is_skew_symmetric(mat=point, atol=atol)
        return has_right_shape

    def random_point(self, n_samples=1, bound=1.0):
        """Sample from a uniform distribution in a cube and project to skew-symmetric.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample each entry.
            Optional, default: 1.

        Returns
        -------
        point : array-like, shape=[..., n, n]
            Sample.
        """
        return self.projection(super().random_point(n_samples, bound))

    @classmethod
    def projection(cls, mat):
        r"""Compute the skew-symmetric component of a matrix.

        The skew-symmetric part of a matrix :math:`X` is defined by

        .. math::
            (X - X^T) / 2

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        skew_sym : array-like, shape=[..., n, n]
            Skew-symmetric matrix.
        """
        return Matrices.to_skew_symmetric(mat)

    def basis_representation(self, matrix_representation):
        """Calculate the coefficients of given matrix in the basis.

        Compute a 1d-array that corresponds to the input matrix in the basis
        representation.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        basis_representation : array-like, shape=[..., dim]
            Representation in the basis.
        """
        if self.n == 2:
            return matrix_representation[..., 1, 0][..., None]
        if self.n == 3:
            vec = gs.stack(
                [
                    matrix_representation[..., 2, 1],
                    matrix_representation[..., 0, 2],
                    matrix_representation[..., 1, 0],
                ]
            )
            return gs.transpose(vec)

        return gs.triu_to_vec(matrix_representation, k=1)



"""Visualization for Geometric Statistics."""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA

import geomstats.backend as gs
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.pre_shape import PreShapeSpace

M32 = Matrices(m=3, n=2)

S32 = PreShapeSpace(k_landmarks=3, ambient_dim=2)
S32.equip_with_group_action("rotations")
S32.equip_with_quotient()

M33 = Matrices(m=3, n=3)
S33 = PreShapeSpace(k_landmarks=3, ambient_dim=3)
S33.equip_with_group_action("rotations")
S33.equip_with_quotient()


class KendallSphere:
    """Class used to plot points in Kendall shape space of 2D triangles.

    David G. Kendall showed that the shape space of 2D triangles is isometric
    to the 2-sphere of radius 1/2 [K1984]_. This class encodes this isometric
    representation, offering a 3D visualization of Kendall shape space of order
    (3,2), and its related objects.

    Attributes
    ----------
    points : list
        List of points to plot on the Kendall sphere.
    coords_type : str
        Type of the points. Can be either 'pre-shape' (for points in Kendall
        pre-shape space) or 'extrinsic' (for points given as 3x2 matrices).
        Optional, default: 'pre-shape'.
    pole : array-like, shape=[3,2]
        Equilateral triangle (north pole).
    ua : array-like, shape=[3,2]
        Tangent vector toward isosceles triangle at vertex A.
    ub : array-like, shape=[3,2]
        Tangent vector toward isosceles triangle at vertex B.
    na : array-like, shape=[3,2]
        Tangent vector such that (ua,na) is a positively oriented
        orthonormal basis of the horizontal space at north pole.

    References
    ----------
    .. [K1984] David G. Kendall. "Shape Manifolds, Procrustean Metrics, and
        Complex Projective Spaces." Bulletin of the London Mathematical
        Society, Volume 16, Issue 2, March 1984, Pages 81–121.
        https://doi.org/10.1112/blms/16.2.81
    """

    def __init__(self, points=None, coords_type="pre-shape"):
        self.points = []
        self.coords_type = coords_type
        self.ax = None
        self.elev, self.azim = None, None

        self.pole = gs.array(
            [[1.0, 0.0], [-0.5, gs.sqrt(3.0) / 2.0], [-0.5, -gs.sqrt(3.0) / 2.0]]
        ) / gs.sqrt(3.0)

        self.ua = gs.array(
            [[-1.0, 0.0], [0.5, gs.sqrt(3.0) / 2.0], [0.5, -gs.sqrt(3.0) / 2.0]]
        ) / gs.sqrt(3.0)

        self.ub = gs.array(
            [[0.5, gs.sqrt(3.0) / 2.0], [0.5, -gs.sqrt(3.0) / 2], [-1.0, 0.0]]
        ) / gs.sqrt(3.0)

        self.na = self.ub - S32.metric.inner_product(self.ub, self.ua) * self.ua
        self.na = self.na / S32.metric.norm(self.na)

        if points is not None:
            self.add_points(points)

    def set_ax(self, ax=None):
        """Set axis."""
        if ax is None:
            ax = plt.subplot(111, projection="3d")

        ax_s = 0.5
        plt.setp(
            ax,
            xlim=(-ax_s, ax_s),
            ylim=(-ax_s, ax_s),
            zlim=(-ax_s, ax_s),
            xlabel="X",
            ylabel="Y",
            zlabel="Z",
        )
        self.ax = ax

    def set_view(self, elev=60.0, azim=0.0):
        """Set azimuth and elevation angle."""
        if self.ax is None:
            self.set_ax()

        self.elev, self.azim = gs.pi * elev / 180, gs.pi * azim / 180
        self.ax.view_init(elev, azim)

    def convert_to_polar_coordinates(self, points):
        """Assign polar coordinates to given pre-shapes."""
        aligned_points = S32.fiber_bundle.align(points, self.pole)
        speeds = S32.metric.log(aligned_points, self.pole)

        coords_theta = gs.arctan2(
            S32.metric.inner_product(speeds, self.na),
            S32.metric.inner_product(speeds, self.ua),
        )
        coords_phi = 2.0 * S32.metric.dist(self.pole, aligned_points)

        return coords_theta, coords_phi

    def convert_to_spherical_coordinates(self, points):
        """Convert polar coordinates to spherical one."""
        coords_theta, coords_phi = self.convert_to_polar_coordinates(points)
        coords_x = 0.5 * gs.cos(coords_theta) * gs.sin(coords_phi)
        coords_y = 0.5 * gs.sin(coords_theta) * gs.sin(coords_phi)
        coords_z = 0.5 * gs.cos(coords_phi)
        spherical_coords = gs.transpose(gs.stack((coords_x, coords_y, coords_z)))
        return spherical_coords

    def add_points(self, points):
        """Add points to draw on the Kendall sphere."""
        if self.coords_type == "extrinsic":
            if not gs.all(M32.belongs(points)):
                raise ValueError("Points do not belong to Matrices(3, 2).")
            points = S32.projection(points)
        elif self.coords_type == "pre-shape" and not gs.all(S32.belongs(points)):
            raise ValueError("Points do not belong to the pre-shape space.")
        points = self.convert_to_spherical_coordinates(points)
        if not isinstance(points, list):
            if points.shape == (3,):
                points = [gs.array(points)]
            else:
                points = list(points)
        self.points.extend(points)

    def clear_points(self):
        """Clear the points to draw."""
        self.points = []

    def draw(self, n_theta=25, n_phi=13, scale=0.05, elev=60.0, azim=0.0):
        """Draw the sphere regularly sampled with corresponding triangles."""
        self.set_ax()
        self.set_view(elev=elev, azim=azim)
        self.ax.set_axis_off()
        plt.tight_layout()

        coords_theta = gs.linspace(0.0, 2.0 * gs.pi, n_theta)
        coords_phi = gs.linspace(0.0, gs.pi, n_phi)

        coords_x = gs.to_numpy(0.5 * gs.outer(gs.sin(coords_phi), gs.cos(coords_theta)))
        coords_y = gs.to_numpy(0.5 * gs.outer(gs.sin(coords_phi), gs.sin(coords_theta)))
        coords_z = gs.to_numpy(
            0.5 * gs.outer(gs.cos(coords_phi), gs.ones_like(coords_theta))
        )

        self.ax.plot_surface(
            coords_x,
            coords_y,
            coords_z,
            rstride=1,
            cstride=1,
            color="grey",
            linewidth=0,
            alpha=0.1,
            zorder=-1,
        )
        self.ax.plot_wireframe(
            coords_x,
            coords_y,
            coords_z,
            linewidths=0.6,
            color="grey",
            alpha=0.6,
            zorder=-1,
        )

        def lim(theta):
            return (
                gs.pi
                - self.elev
                + (2.0 * self.elev - gs.pi) / gs.pi * abs(self.azim - theta)
            )

        for theta in gs.linspace(0.0, 2.0 * gs.pi, n_theta // 2 + 1):
            for phi in gs.linspace(0.0, gs.pi, n_phi):
                if theta <= self.azim + gs.pi and phi <= lim(theta):
                    self.draw_triangle(theta, phi, scale)
                if theta > self.azim + gs.pi and phi < lim(
                    2.0 * self.azim + 2.0 * gs.pi - theta
                ):
                    self.draw_triangle(theta, phi, scale)

    def draw_triangle(self, theta, phi, scale):
        """Draw the corresponding triangle on the sphere at theta, phi."""
        u_theta = gs.cos(theta) * self.ua + gs.sin(theta) * self.na
        triangle = gs.cos(phi / 2.0) * self.pole + gs.sin(phi / 2.0) * u_theta
        triangle = scale * triangle
        triangle3d = gs.transpose(
            gs.stack((triangle[:, 0], triangle[:, 1], 0.5 * gs.ones(3)))
        )
        triangle3d = self.rotation(theta, phi) @ gs.transpose(triangle3d)

        x = list(triangle3d[0]) + [triangle3d[0, 0]]
        y = list(triangle3d[1]) + [triangle3d[1, 0]]
        z = list(triangle3d[2]) + [triangle3d[2, 0]]

        self.ax.plot3D(x, y, z, "grey", zorder=1)
        c = ["red", "green", "blue"]
        for i in range(3):
            self.ax.scatter(x[i], y[i], z[i], color=c[i], s=10, alpha=1, zorder=1)

    @staticmethod
    def rotation(theta, phi):
        """Rotation sending a triangle at pole to location theta, phi."""
        rot_th = gs.array(
            [
                [gs.cos(theta), -gs.sin(theta), 0.0],
                [gs.sin(theta), gs.cos(theta), 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        rot_phi = gs.array(
            [
                [gs.cos(phi), 0.0, gs.sin(phi)],
                [0.0, 1.0, 0.0],
                [-gs.sin(phi), 0, gs.cos(phi)],
            ]
        )
        return rot_th @ rot_phi @ gs.transpose(rot_th)

    def draw_points(self, alpha=1, zorder=0, **kwargs):
        """Draw points on the Kendall sphere."""
        points_x = [gs.to_numpy(point)[0] for point in self.points]
        points_y = [gs.to_numpy(point)[1] for point in self.points]
        points_z = [gs.to_numpy(point)[2] for point in self.points]
        self.ax.scatter(
            points_x, points_y, points_z, alpha=alpha, zorder=zorder, **kwargs
        )

    def draw_curve(self, alpha=1, zorder=0, **kwargs):
        """Draw a curve on the Kendall sphere."""
        points_x = [gs.to_numpy(point)[0] for point in self.points]
        points_y = [gs.to_numpy(point)[1] for point in self.points]
        points_z = [gs.to_numpy(point)[2] for point in self.points]
        self.ax.plot3D(
            points_x, points_y, points_z, alpha=alpha, zorder=zorder, **kwargs
        )

    def draw_vector(self, tangent_vec, base_point, **kwargs):
        """Draw one vector in the tangent space to sphere at a base point."""
        norm = S32.quotient.metric.norm(tangent_vec, base_point)
        exp = S32.quotient.metric.exp(tangent_vec, base_point)
        bp = self.convert_to_spherical_coordinates(base_point)
        exp = self.convert_to_spherical_coordinates(exp)
        tv = exp - gs.dot(exp, 2.0 * bp) * 2.0 * bp
        tv = tv / gs.linalg.norm(tv) * norm
        self.ax.quiver(bp[0], bp[1], bp[2], tv[0], tv[1], tv[2], **kwargs)


class KendallDisk:
    """Class used to plot points in Kendall shape space of 3D triangles.

    The shape space of 2D triangles is isometric to the 2-sphere of radius 1/2
    [K1984]_. This isometry induced another isometry between the shape space of
    3D triangle and the 1-ball of radius pi/4 [LK1993]_. Following the first
    visualization class "KendallSphere" for 2D triangles, this class encodes
    the 2D isometric representation of Kendall shape space of order (3,3).

    Attributes
    ----------
    points : list
        List of points to plot on the Kendall sphere.
    coords_type : str
        Type of the points. Can be either 'pre-shape' (for points in Kendall
        pre-shape space) or 'extrinsic' (for points given as 3x2 matrices).
        Optional, default: 'pre-shape'.
    pole : array-like, shape=[3,2]
        Equilateral triangle in 2D (north pole).
    centre : array-like, shape=[3,3]
        Equilateral triangle in 3D (centre).
    ua : array-like, shape=[3,2]
        Tangent vector at north pole toward isosceles triangle at vertex A.
    ub : array-like, shape=[3,2]
        Tangent vector at north pole toward isosceles triangle at vertex B.
    na : array-like, shape=[3,2]
        Tangent vector such that (ua,na) is a positively oriented
        orthonormal basis of the horizontal space at north pole.

    References
    ----------
    .. [K1984] David G. Kendall. "Shape Manifolds, Procrustean Metrics, and
        Complex Projective Spaces." Bulletin of the London Mathematical
        Society, Volume 16, Issue 2, March 1984, Pages 81–121.
        https://doi.org/10.1112/blms/16.2.81
    .. [LK1993] Huiling Le and David G. Kendall. "The Riemannian structure of
        Euclidean shape spaces: a novel environment for statistics." Annals of
        statistics, 1993, vol. 21, no 3, p. 1225-1271.
        https://doi.org/10.1112/blms/16.2.81
    """

    def __init__(self, points=None, coords_type="pre-shape"):
        self.points = []
        self.coords_type = coords_type
        self.ax = None

        self.pole = gs.array(
            [[1.0, 0.0], [-0.5, gs.sqrt(3.0) / 2.0], [-0.5, -gs.sqrt(3.0) / 2.0]]
        ) / gs.sqrt(3.0)

        self.centre = gs.array(
            [
                [1.0, 0.0, 0.0],
                [-0.5, gs.sqrt(3.0) / 2.0, 0.0],
                [-0.5, -gs.sqrt(3.0) / 2.0, 0.0],
            ]
        ) / gs.sqrt(3.0)

        self.ua = gs.array(
            [[-1.0, 0.0], [0.5, gs.sqrt(3.0) / 2.0], [0.5, -gs.sqrt(3.0) / 2.0]]
        ) / gs.sqrt(3.0)

        self.ub = gs.array(
            [[0.5, gs.sqrt(3.0) / 2.0], [0.5, -gs.sqrt(3.0) / 2], [-1.0, 0.0]]
        ) / gs.sqrt(3.0)

        self.na = self.ub - S32.metric.inner_product(self.ub, self.ua) * self.ua
        self.na = self.na / S32.metric.norm(self.na)

        if points is not None:
            self.add_points(points)

    def set_ax(self, ax=None):
        """Set axis."""
        if ax is None:
            ax = plt.subplot()

        ax_s = gs.pi / 4 + 0.05
        plt.setp(ax, xlim=(-ax_s, ax_s), ylim=(-ax_s, ax_s), xlabel="X", ylabel="Y")
        self.ax = ax

    def convert_to_polar_coordinates(self, points):
        """Assign polar coordinates to given pre-shapes."""
        aligned_points = S33.fiber_bundle.align(points, self.centre)
        aligned_points2d = aligned_points[..., :, :2]
        speeds = S32.metric.log(aligned_points2d, self.pole)

        coords_r = S32.metric.dist(self.pole, aligned_points2d)
        coords_theta = gs.arctan2(
            S32.metric.inner_product(speeds, self.na),
            S32.metric.inner_product(speeds, self.ua),
        )

        return coords_r, coords_theta

    def convert_to_planar_coordinates(self, points):
        """Convert polar coordinates to spherical one."""
        coords_r, coords_theta = self.convert_to_polar_coordinates(points)
        coords_x = coords_r * gs.cos(coords_theta)
        coords_y = coords_r * gs.sin(coords_theta)
        planar_coords = gs.transpose(gs.stack((coords_x, coords_y)))
        return planar_coords

    def add_points(self, points):
        """Add points to draw on the Kendall disk."""
        if self.coords_type == "extrinsic":
            if not gs.all(M33.belongs(points)):
                raise ValueError("Points do not belong to Matrices(3, 3).")
            points = S33.projection(points)
        elif self.coords_type == "pre-shape" and not gs.all(S33.belongs(points)):
            raise ValueError("Points do not belong to the pre-shape space.")
        points = self.convert_to_planar_coordinates(points)
        if not isinstance(points, list):
            if points.shape == (2,):
                points = [gs.array(points)]
            else:
                points = list(points)
        self.points.extend(points)

    def clear_points(self):
        """Clear the points to draw."""
        self.points = []

    def draw(self, n_r=7, n_theta=25, scale=0.05):
        """Draw the disk regularly sampled with corresponding triangles."""
        self.set_ax()
        self.ax.set_axis_off()
        plt.tight_layout()

        coords_r = gs.linspace(0.0, gs.pi / 4.0, n_r)
        coords_theta = gs.linspace(0.0, 2.0 * gs.pi, n_theta)

        coords_x = gs.to_numpy(gs.outer(coords_r, gs.cos(coords_theta)))
        coords_y = gs.to_numpy(gs.outer(coords_r, gs.sin(coords_theta)))

        self.ax.fill(
            list(coords_x[-1, :]),
            list(coords_y[-1, :]),
            color="grey",
            alpha=0.1,
            zorder=-1,
        )
        for i_r in range(n_r):
            self.ax.plot(
                coords_x[i_r, :],
                coords_y[i_r, :],
                linewidth=0.6,
                color="grey",
                alpha=0.6,
                zorder=-1,
            )
        for i_t in range(n_theta):
            self.ax.plot(
                coords_x[:, i_t],
                coords_y[:, i_t],
                linewidth=0.6,
                color="grey",
                alpha=0.6,
                zorder=-1,
            )

        for r in gs.linspace(0.0, gs.pi / 4, n_r):
            for theta in gs.linspace(0.0, 2.0 * gs.pi, n_theta // 2 + 1):
                if theta == 0.0:
                    self.draw_triangle(0.0, 0.0, scale)
                else:
                    self.draw_triangle(r, theta, scale)

    def draw_triangle(self, r, theta, scale):
        """Draw the corresponding triangle on the disk at r, theta."""
        u_theta = gs.cos(theta) * self.ua + gs.sin(theta) * self.na
        triangle = gs.cos(r) * self.pole + gs.sin(r) * u_theta
        triangle = scale * triangle

        x = list(r * gs.cos(theta) + triangle[:, 0])
        x = x + [x[0]]
        y = list(r * gs.sin(theta) + triangle[:, 1])
        y = y + [y[0]]

        self.ax.plot(x, y, "grey", zorder=1)
        c = ["red", "green", "blue"]
        for i in range(3):
            self.ax.scatter(x[i], y[i], color=c[i], s=10, alpha=1, zorder=1)

    def draw_points(self, alpha=1, zorder=0, **kwargs):
        """Draw points on the Kendall disk."""
        points_x = [gs.to_numpy(point)[0] for point in self.points]
        points_y = [gs.to_numpy(point)[1] for point in self.points]
        self.ax.scatter(points_x, points_y, alpha=alpha, zorder=zorder, **kwargs)

    def draw_curve(self, alpha=1, zorder=0, **kwargs):
        """Draw a curve on the Kendall disk."""
        points_x = [gs.to_numpy(point)[0] for point in self.points]
        points_y = [gs.to_numpy(point)[1] for point in self.points]
        self.ax.plot(points_x, points_y, alpha=alpha, zorder=zorder, **kwargs)

    def draw_vector(self, tangent_vec, base_point, tol=1e-03, **kwargs):
        """Draw one vector in the tangent space to disk at a base point."""
        r_bp, th_bp = self.convert_to_polar_coordinates(base_point)
        bp = gs.array(
            [
                gs.cos(th_bp) * gs.sin(2 * r_bp),
                gs.sin(th_bp) * gs.sin(2 * r_bp),
                gs.cos(2 * r_bp),
            ]
        )
        r_exp, th_exp = self.convert_to_polar_coordinates(
            S33.quotient.metric.exp(
                tol * tangent_vec / S33.quotient.metric.norm(tangent_vec, base_point),
                base_point,
            )
        )
        exp = gs.array(
            [
                gs.cos(th_exp) * gs.sin(2 * r_exp),
                gs.sin(th_exp) * gs.sin(2 * r_exp),
                gs.cos(2 * r_exp),
            ]
        )
        pole = gs.array([0.0, 0.0, 1.0])

        tv = exp - gs.dot(exp, bp) * bp
        u_tv = tv / gs.linalg.norm(tv)
        u_r = (gs.dot(pole, bp) * bp - pole) / gs.linalg.norm(
            gs.dot(pole, bp) * bp - pole
        )
        u_th = gs.cross(bp, u_r)
        x_r, x_th = gs.dot(u_tv, u_r), gs.dot(u_tv, u_th)

        bp = self.convert_to_planar_coordinates(base_point)
        u_r = bp / gs.linalg.norm(bp)
        u_th = gs.array([[0.0, -1.0], [1.0, 0.0]]) @ u_r
        tv = S33.quotient.metric.norm(tangent_vec, base_point) * (
            x_r * u_r + x_th * u_th
        )

        self.ax.quiver(bp[0], bp[1], tv[0], tv[1], **kwargs)



"""Affine connections.

Lead author: Nicolas Guigui.
"""

from abc import ABC

import geomstats.backend as gs
import geomstats.errors


def _check_log_solver(connection, raise_=True):
    if not hasattr(connection, "log_solver"):
        if raise_:
            raise ValueError(
                "Requires `self.log_solver`. "
                "Check `geomstats.numerics.geodesic` for available solvers."
            )
        return False
    return True


def _check_exp_solver(connection, raise_=True):
    if not hasattr(connection, "exp_solver"):
        if raise_:
            raise ValueError(
                "Requires `self.exp_solver`. "
                "Check `geomstats.numerics.geodesic` for available solvers."
            )
        return False
    return True


class Connection(ABC):
    r"""Class for affine connections.

    Parameters
    ----------
    space : Manifold object
        M in the tuple (M, g).
    """

    def __init__(self, space):
        self._space = space

    def christoffels(self, base_point=None):
        """Christoffel symbols associated with the connection.

        The contravariant index is on the first dimension.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        gamma : array-like, shape=[..., dim, dim, dim]
            Christoffel symbols, with the contravariant index on
            the first dimension.
        """
        raise NotImplementedError("The Christoffel symbols are not implemented.")

    def geodesic_equation(self, state):
        """Compute the geodesic ODE associated with the connection.

        Parameters
        ----------
        state : array-like, shape=[..., 2, dim]
            Tangent vector at the position.

        Returns
        -------
        geodesic_ode : array-like, shape=[..., 2, dim]
            Value of the vector field to be integrated at position.
        """
        if self._space.point_ndim != 1:
            raise NotImplementedError(
                "The geodesic equation is only implemented for points "
                "represented by a vector."
            )
        position = state[..., 0, :]
        velocity = state[..., 1, :]
        gamma = self.christoffels(position)
        equation = gs.einsum("...kij,...i->...kj", gamma, velocity)
        equation = -gs.einsum("...kj,...j->...k", equation, velocity)
        return gs.stack([velocity, equation], axis=-2)

    def exp(self, tangent_vec, base_point):
        """Exponential map associated to the affine connection.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        exp : array-like, shape=[..., dim]
            Point on the manifold.
        """
        _check_exp_solver(self)
        return self.exp_solver.exp(tangent_vec, base_point)

    def log(self, point, base_point):
        """Compute logarithm map associated to the affine connection.

        Solve the boundary value problem associated to the geodesic equation
        using the Christoffel symbols and conjugate gradient descent.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point on the manifold.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.
        n_steps : int
            Number of discrete time steps to take in the integration.
            Optional, default: N_STEPS.
        step : str, {'euler', 'rk4'}
            Numerical scheme to use for integration.
            Optional, default: 'euler'.
        max_iter
        verbose
        tol

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at the base point.
        """
        _check_log_solver(self)
        return self.log_solver.log(point, base_point)

    def _pole_ladder_step(
        self, base_point, next_point, base_shoot, return_geodesics=False
    ):
        """Compute one Pole Ladder step.

        One step of pole ladder scheme [LP2013a]_ using the geodesic to
        transport along as main_geodesic of the parallelogram.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold, from which to transport.
        next_point : array-like, shape=[..., dim]
            Point on the manifold, to transport to.
        base_shoot : array-like, shape=[..., dim]
            Point on the manifold, end point of the geodesics starting
            from the base point with initial speed to be transported.
        return_geodesics : bool, optional (defaults to False)
            Whether to return the geodesics of the
            construction.

        Returns
        -------
        next_step : dict of array-like and callable with following keys:
            next_tangent_vec : array-like, shape=[..., dim]
                Tangent vector at end point.
            end_point : array-like, shape=[..., dim]
                Point on the manifold, closes the geodesic parallelogram of the
                construction.
            geodesics : list of callable, len=3 (only if
            `return_geodesics=True`)
                Three geodesics of the construction.

        References
        ----------
        .. [LP2013a] Marco Lorenzi, Xavier Pennec. Efficient Parallel Transport
            of Deformations in Time Series of Images: from Schild's to
            Pole Ladder. Journal of Mathematical Imaging and Vision, Springer
            Verlag, 2013,50 (1-2), pp.5-17. ⟨10.1007/s10851-013-0470-3⟩
        """
        mid_tangent_vector_to_shoot = (
            1.0 / 2.0 * self.log(base_point=base_point, point=next_point)
        )

        mid_point = self.exp(
            base_point=base_point, tangent_vec=mid_tangent_vector_to_shoot
        )

        tangent_vector_to_shoot = -self.log(base_point=mid_point, point=base_shoot)

        end_shoot = self.exp(base_point=mid_point, tangent_vec=tangent_vector_to_shoot)

        geodesics = []
        if return_geodesics:
            main_geodesic = self.geodesic(
                initial_point=base_point, end_point=next_point
            )
            diagonal = self.geodesic(initial_point=mid_point, end_point=base_shoot)
            final_geodesic = self.geodesic(
                initial_point=next_point, end_point=end_shoot
            )
            geodesics = [main_geodesic, diagonal, final_geodesic]
        return {"geodesics": geodesics, "end_point": end_shoot}

    def _schild_ladder_step(
        self, base_point, next_point, base_shoot, return_geodesics=False
    ):
        """Compute one Schild's Ladder step.

        One step of the Schild's ladder scheme [LP2013a]_ using the geodesic to
        transport along as one side of the parallelogram.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold, from which to transport.
        next_point : array-like, shape=[..., dim]
            Point on the manifold, to transport to.
        base_shoot : array-like, shape=[..., dim]
            Point on the manifold, end point of the geodesics starting
            from the base point with initial speed to be transported.
        return_geodesics : bool
            Whether to return points computed along each geodesic of the
            construction.
            Optional, default: False.

        Returns
        -------
        transported_tangent_vector : array-like, shape=[..., dim]
            Tangent vector at end point.
        end_point : array-like, shape=[..., dim]
            Point on the manifold, closes the geodesic parallelogram of the
            construction.

        References
        ----------
        .. [LP2013a] Marco Lorenzi, Xavier Pennec. Efficient Parallel Transport
            of Deformations in Time Series of Images: from Schild's to
            Pole Ladder. Journal of Mathematical Imaging and Vision, Springer
            Verlag, 2013,50 (1-2), pp.5-17. ⟨10.1007/s10851-013-0470-3⟩
        """
        mid_tangent_vector_to_shoot = (
            1.0 / 2.0 * self.log(base_point=base_shoot, point=next_point)
        )

        mid_point = self.exp(
            base_point=base_shoot, tangent_vec=mid_tangent_vector_to_shoot
        )

        tangent_vector_to_shoot = -self.log(base_point=mid_point, point=base_point)

        end_shoot = self.exp(base_point=mid_point, tangent_vec=tangent_vector_to_shoot)

        geodesics = []
        if return_geodesics:
            main_geodesic = self.geodesic(
                initial_point=base_point, end_point=next_point
            )
            diagonal = self.geodesic(initial_point=base_point, end_point=end_shoot)
            second_diagonal = self.geodesic(
                initial_point=base_shoot, end_point=next_point
            )
            final_geodesic = self.geodesic(
                initial_point=next_point, end_point=end_shoot
            )
            geodesics = [main_geodesic, diagonal, second_diagonal, final_geodesic]
        return {"geodesics": geodesics, "end_point": end_shoot}

    def ladder_parallel_transport(
        self,
        tangent_vec,
        base_point,
        direction,
        n_rungs=1,
        scheme="pole",
        alpha=1,
        return_geodesics=False,
    ):
        """Approximate parallel transport using the pole ladder scheme.

        Approximate Parallel transport using either the pole ladder or the
        Schild's ladder scheme [LP2013b]_. Pole ladder is exact in symmetric
        spaces and of order two in general while Schild's ladder is a first
        order approximation [GP2020]_. Both schemes are available on any affine
        connection manifolds whose exponential and logarithm maps are
        implemented. `tangent_vec` is transported along the geodesic starting
        at the base_point with initial tangent vector `direction`.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point to transport.
        direction : array-like, shape=[..., dim]
            Tangent vector at base point, initial speed of the geodesic along
            which to transport.
        base_point : array-like, shape=[..., dim]
            Point on the manifold, initial position of the geodesic along
            which to transport.
        n_rungs : int
            Number of steps of the ladder.
            Optional, default: 1.
        scheme : str, {'pole', 'schild'}
            The scheme to use for the construction of the ladder at each step.
            Optional, default: 'pole'.
        alpha : float
            Exponent for the scaling of the vector to transport. Must be
            greater or equal to 1, 2 is optimal. See [GP2020]_.
            Optional, default: 2

        Returns
        -------
        ladder : dict of array-like and callable with following keys
            transported_tangent_vector : array-like, shape=[..., dim]
                Approximation of the parallel transport of tangent vector a.
            trajectory : list of list of callable, len=n_steps
                List of lists containing the geodesics of the
                construction, only if `return_geodesics=True` in the step
                function. The geodesics are methods of the class connection.

        References
        ----------
        .. [LP2013b] Lorenzi, Marco, and Xavier Pennec. “Efficient Parallel
            Transport of Deformations in Time Series of Images: From Schild to
            Pole Ladder.” Journal of Mathematical Imaging and Vision 50, no. 1
            (September 1, 2014): 5–17.
            https://doi.org/10.1007/s10851-013-0470-3.

        .. [GP2020] Guigui, Nicolas, and Xavier Pennec. “Numerical Accuracy
            of Ladder Schemes for Parallel Transport on Manifolds.”
            Foundations of Computational Mathematics, June 18, 2021.
            https://doi.org/10.1007/s10208-021-09515-x.
        """
        geomstats.errors.check_integer(n_rungs, "n_rungs")
        if alpha < 1:
            raise ValueError("alpha must be greater or equal to one")
        current_point = base_point
        next_tangent_vec = tangent_vec / (n_rungs**alpha)
        methods = {"pole": self._pole_ladder_step, "schild": self._schild_ladder_step}
        single_step = methods[scheme]
        base_shoot = self.exp(base_point=current_point, tangent_vec=next_tangent_vec)
        trajectory = []
        for i_point in range(n_rungs):
            frac_tan_vector_b = (i_point + 1) / n_rungs * direction
            next_point = self.exp(base_point=base_point, tangent_vec=frac_tan_vector_b)
            next_step = single_step(
                base_point=current_point,
                next_point=next_point,
                base_shoot=base_shoot,
                return_geodesics=return_geodesics,
            )
            current_point = next_point
            base_shoot = next_step["end_point"]
            trajectory.append(next_step["geodesics"])
        transported_tangent_vec = self.log(base_shoot, current_point)
        if n_rungs % 2 == 1 and scheme == "pole":
            transported_tangent_vec *= -1.0
        transported_tangent_vec *= n_rungs**alpha
        return {
            "transported_tangent_vec": transported_tangent_vec,
            "end_point": current_point,
            "trajectory": trajectory,
        }

    def riemann_tensor(self, base_point=None):
        r"""Compute Riemannian tensor at base_point.

        In the literature the Riemannian curvature tensor is noted :math:`R_{ijk}^l`.

        Following tensor index convention (ref. Wikipedia), we have:
        :math:`R_{ijk}^l = dx^l(R(X_j, X_k)X_i)`

        which gives :math:`R_{ijk}^l` as a sum of four terms:

        .. math::
            \partial_j \Gamma^l_{ki} - \partial_k \Gamma^l_{ji}
            + \Gamma^l_{jm} \Gamma^m_{ki} - \Gamma^l_{km} \Gamma^m_{ji}

        Note that geomstats puts the contravariant index on
        the first dimension of the Christoffel symbols.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        riemann_curvature : array-like, shape=[..., dim, dim, dim, dim]
            riemann_tensor[...,i,j,k,l] = R_{ijk}^l
            Riemannian tensor curvature,
            with the contravariant index on the last dimension.
        """
        if len(self._space.shape) > 1:
            raise NotImplementedError(
                "Riemann tensor not implemented for manifolds with points of ndim > 1."
            )
        christoffels = self.christoffels(base_point)
        jacobian_christoffels = gs.autodiff.jacobian_vec(self.christoffels)(base_point)

        prod_christoffels = gs.einsum(
            "...ijk,...klm->...ijlm", christoffels, christoffels
        )
        riemann_curvature = (
            gs.einsum("...ijlm->...lmji", jacobian_christoffels)
            - gs.einsum("...ijlm->...ljmi", jacobian_christoffels)
            + gs.einsum("...ijlm->...mjli", prod_christoffels)
            - gs.einsum("...ijlm->...lmji", prod_christoffels)
        )

        return riemann_curvature

    def curvature(self, tangent_vec_a, tangent_vec_b, tangent_vec_c, base_point=None):
        r"""Compute the Riemann curvature map R.

        For three tangent vectors at base point :math:`P`:

        - :math:`X|_P = tangent\_vec\_a`,
        - :math:`Y|_P = tangent\_vec\_b`,
        - :math:`Z|_P = tangent\_vec\_c`,

        the curvature(X, Y, Z, P) is defined by
        :math:`R(X,Y)Z = \nabla_X \nabla_Y Z - \nabla_Y \nabla_X Z - \nabla_[X, Y]Z`.

        The output is the tangent vector:
        :math:`dx^l(R(X, Y)Z) = R_{ijk}^l X_j Y_k Z_i`
        written with Einstein notation.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        tangent_vec_b : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        tangent_vec_c : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        base_point :  array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        curvature : array-like, shape=[..., dim]
            curvature(X, Y, Z, P)[..., l] = dx^l(R(X, Y)Z)
            Tangent vector at `base_point`.
        """
        riemann = self.riemann_tensor(base_point)
        curvature = gs.einsum(
            "...ijkl, ...j, ...k, ...i -> ...l",
            riemann,
            tangent_vec_a,
            tangent_vec_b,
            tangent_vec_c,
        )
        return curvature

    def ricci_tensor(self, base_point=None):
        r"""Compute Ricci curvature tensor at base_point.

        The Ricci curvature tensor :math:`\mathrm{Ric}_{ij}` is defined as:
        :math:`\mathrm{Ric}_{ij} = R_{ikj}^k`
        with Einstein notation.

        Parameters
        ----------
        base_point :  array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        ricci_tensor : array-like, shape=[..., dim, dim]
            ricci_tensor[...,i,j] = Ric_{ij}
            Ricci tensor curvature.
        """
        riemann_tensor = self.riemann_tensor(base_point)
        ricci_tensor = gs.einsum("...ijkj -> ...ik", riemann_tensor)
        return ricci_tensor

    def directional_curvature(self, tangent_vec_a, tangent_vec_b, base_point=None):
        r"""Compute the directional curvature (tidal force operator).

        For two tangent vectors at base_point :math:`P`:

        - :math:`X|_P = tangent\_vec\_a`,
        - :math:`Y|_P = tangent\_vec\_b`,

        the directional curvature, better known
        in relativity as the tidal force operator,
        is defined by
        :math:`R_Y(X) = R(Y,X)Y`.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        tangent_vec_b : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        base_point :  array-like, shape=[..., dim]
            Base-point on the manifold.

        Returns
        -------
        directional_curvature : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        """
        return self.curvature(tangent_vec_b, tangent_vec_a, tangent_vec_b, base_point)

    def curvature_derivative(
        self,
        tangent_vec_a,
        tangent_vec_b,
        tangent_vec_c,
        tangent_vec_d,
        base_point=None,
    ):
        r"""Compute the covariant derivative of the curvature.

        For four tangent vectors at base_point :math:`P`:

        - :math:`H|_P = tangent\_vec\_a`,
        - :math:`X|_P = tangent\_vec\_b`,
        - :math:`Y|_P = tangent\_vec\_c`,
        - :math:`Z|_P = tangent\_vec\_d`,

        the covariant derivative of the curvature is defined as:
        :math:`(\nabla_H R)(X, Y) Z |_P`.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        tangent_vec_b : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        tangent_vec_c : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        tangent_vec_d : array-like, shape=[..., dim]
            Tangent vector at `base_point`.
        base_point :  array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        curvature_derivative : array-like, shape=[..., dim]
            Tangent vector at base-point.
        """
        raise NotImplementedError(
            "The covariant derivative of the curvature is not implemented."
        )

    def directional_curvature_derivative(
        self, tangent_vec_a, tangent_vec_b, base_point=None
    ):
        r"""Compute the covariant derivative of the directional curvature.

        For tangent vector fields at base_point :math:`P`:

        - :math:`X|_P = tangent\_vec\_a`,
        - :math:`Y|_P = tangent\_vec\_b`,

        the covariant derivative (in the direction `X`)
        :math:`(\nabla_X R_Y)(X) |_P = (\nabla_X R)(Y, X) Y |_P` of the
        directional curvature (in the direction `Y`)
        :math:`R_Y(X) = R(Y, X) Y`
        is a quadratic tensor in `X` and `Y` that
        plays an important role in the computation of the moments of the
        empirical Fréchet mean.

        References
        ----------
        .. [Pennec] Pennec, Xavier. Curvature effects on the empirical mean in
            Riemannian and affine Manifolds: a non-asymptotic high
            concentration expansion in the small-sample regime. Preprint. 2019.
            https://arxiv.org/abs/1906.07418
        """
        return self.curvature_derivative(
            tangent_vec_a, tangent_vec_b, tangent_vec_a, tangent_vec_b, base_point
        )

    def _geodesic_from_exp(self, initial_point, initial_tangent_vec):
        """Generate parameterized function for the geodesic curve.

        Parameters
        ----------
        initial_point : array-like, shape=[..., dim]
            Point on the manifold, initial point of the geodesic.
        initial_tangent_vec : array-like, shape=[..., dim],
            Tangent vector at base point, the initial speed of the geodesics.
            Optional, default: None.
            If None, an end point must be given and a logarithm is computed.

        Returns
        -------
        path : callable
            Time parameterized geodesic curve. If a batch of initial
            conditions is passed, the output array's first dimension
            represents the different initial conditions, and the second
            corresponds to time.
        """
        ndim = self._space.point_ndim
        ijk = "ijk"[: self._space.point_ndim]

        multiple_tangent = initial_tangent_vec.ndim > ndim
        multiple_point = initial_point.ndim > ndim
        is_batch = multiple_tangent or multiple_point

        def path(t):
            """Generate parameterized function for geodesic curve.

            Parameters
            ----------
            t : array-like, shape=[n_points,]
                Times at which to compute points of the geodesics.
            """
            t = gs.array(t)
            t = gs.cast(t, initial_tangent_vec.dtype)
            t = gs.to_ndarray(t, to_ndim=1)

            tangent_vecs = gs.einsum(f"n,...{ijk}->...n{ijk}", t, initial_tangent_vec)
            if is_batch:
                if multiple_point and multiple_tangent:
                    points_at_time_t = [
                        self.exp(tv, pt) for tv, pt in zip(tangent_vecs, initial_point)
                    ]

                elif multiple_point:
                    points_at_time_t = [
                        self.exp(tangent_vecs, pt) for pt in initial_point
                    ]

                elif multiple_tangent:
                    points_at_time_t = [
                        self.exp(tv, initial_point) for tv in tangent_vecs
                    ]

                return gs.stack(points_at_time_t, axis=0)

            return self.exp(tangent_vecs, initial_point)

        return path

    def _geodesic_ivp(self, initial_point, initial_tangent_vec):
        """Solve geodesic initial value problem.

        Compute the parameterized function for the geodesic starting at
        initial_point with initial velocity given by initial_tangent_vec.

        Parameters
        ----------
        initial_point : array-like, shape=[..., dim]
            Initial point.

        initial_tangent_vec : array-like, shape=[..., dim]
            Tangent vector at initial point.

        Returns
        -------
        path : function
            Parameterized function for the geodesic curve starting at
            initial_point with velocity initial_tangent_vec.
        """
        if _check_exp_solver(self, raise_=False) and self.exp_solver.solves_ivp:
            return self.exp_solver.geodesic_ivp(initial_tangent_vec, initial_point)

        return self._geodesic_from_exp(initial_point, initial_tangent_vec)

    def _geodesic_bvp(self, initial_point, end_point):
        """Solve geodesic boundary problem.

        Compute the parameterized function for the geodesic starting at
        initial_point and ending at end_point.

        Parameters
        ----------
        initial_point : array-like, shape=[..., dim]
            Initial point.
        end_point : array-like, shape=[..., dim]
            End point.

        Returns
        -------
        path : function
            Parameterized function for the geodesic curve starting at
            initial_point and ending at end_point.
        """
        if _check_log_solver(self, raise_=False) and self.log_solver.solves_bvp:
            return self.log_solver.geodesic_bvp(end_point, initial_point)
        return NotImplemented

    def geodesic(self, initial_point, end_point=None, initial_tangent_vec=None):
        """Generate parameterized function for the geodesic curve.

        Geodesic curve defined by either:

        - an initial point and an initial tangent vector,
        - an initial point and an end point.

        Parameters
        ----------
        initial_point : array-like, shape=[..., dim]
            Point on the manifold, initial point of the geodesic.
        end_point : array-like, shape=[..., dim], optional
            Point on the manifold, end point of the geodesic. If None,
            an initial tangent vector must be given.
        initial_tangent_vec : array-like, shape=[..., dim],
            Tangent vector at base point, the initial speed of the geodesics.
            Optional, default: None.
            If None, an end point must be given and a logarithm is computed.

        Returns
        -------
        path : callable
            Time parameterized geodesic curve. If a batch of initial
            conditions is passed, the output array's first dimension
            represents the different initial conditions, and the second
            corresponds to time.
        """
        if end_point is None and initial_tangent_vec is None:
            raise ValueError(
                "Specify an end point or an initial tangent "
                "vector to define the geodesic."
            )
        if end_point is not None:
            if initial_tangent_vec is not None:
                raise ValueError(
                    "Cannot specify both an end point and an initial tangent vector."
                )

            out = self._geodesic_bvp(initial_point, end_point)
            if out is not NotImplemented:
                return out

            initial_tangent_vec = self.log(end_point, initial_point)

        return self._geodesic_ivp(initial_point, initial_tangent_vec)

    def parallel_transport(
        self, tangent_vec, base_point, direction=None, end_point=None
    ):
        r"""Compute the parallel transport of a tangent vector.

        Closed-form solution for the parallel transport of a tangent vector
        along the geodesic between two points `base_point` and `end_point`
        or alternatively defined by :math:`t \mapsto exp_{(base\_point)}(
        t*direction)`.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {dim, [n, m]}]
            Tangent vector at base point to be transported.
        base_point : array-like, shape=[..., {dim, [n, m]}]
            Point on the manifold. Point to transport from.
        direction : array-like, shape=[..., {dim, [n, m]}]
            Tangent vector at base point, along which the parallel transport
            is computed.
            Optional, default: None.
        end_point : array-like, shape=[..., {dim, [n, m]}]
            Point on the manifold. Point to transport to.
            Optional, default: None.

        Returns
        -------
        transported_tangent_vec: array-like, shape=[..., {dim, [n, m]}]
            Transported tangent vector at `exp_(base_point)(tangent_vec_b)`.
        """
        raise NotImplementedError(
            "The closed-form solution of parallel transport is not known, "
            "use the ladder_parallel_transport instead."
        )

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        This is is the supremum of radii r for which the exponential map is a
        diffeomorphism from the open ball of radius r centered at the base
        point onto its image.

        Parameters
        ----------
        base_point : array-like, shape=[..., {dim, [n, m]}]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        raise NotImplementedError("The injectivity range is not implemented yet.")



import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data


class _ManifoldTestCaseMixins:
    # TODO: check default_coords_type correcteness if intrinsic by comparing
    # with point shape?
    tangent_to_multiple = False

    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    def test_dim(self, expected):
        self.assertEqual(self.space.dim, expected)

    def test_belongs(self, point, expected, atol):
        res = self.space.belongs(point, atol=atol)
        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_not_belongs(self, n_points, atol):
        shape = [val + 1 for val in self.space.shape]

        batch_shape = [n_points] if n_points else []
        point = gs.ones(batch_shape + shape)
        expected = gs.zeros(n_points, dtype=bool)

        self.test_belongs(point, expected, atol)

    @pytest.mark.random
    def test_random_point_belongs(self, n_points, atol):
        """Check random point belongs to manifold.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point = self.data_generator.random_point(n_points)
        expected = gs.ones(n_points, dtype=bool)

        self.test_belongs(point, expected, atol)

    @pytest.mark.shape
    def test_random_point_shape(self, n_points):
        """Check random point shape.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        """
        point = self.data_generator.random_point(n_points)

        expected_ndim = self.space.point_ndim + int(n_points > 1)
        self.assertEqual(gs.ndim(point), expected_ndim)

        self.assertAllEqual(gs.shape(point)[-self.space.point_ndim :], self.space.shape)

        if n_points > 1:
            self.assertEqual(gs.shape(point)[0], n_points)

    def test_is_tangent(self, vector, base_point, expected, atol):
        res = self.space.is_tangent(vector, base_point, atol=atol)
        self.assertAllEqual(res, expected)

    def test_to_tangent(self, vector, base_point, expected, atol):
        tangent_vec = self.space.to_tangent(vector, base_point)
        self.assertAllClose(tangent_vec, expected, atol=atol)

    @pytest.mark.vec
    def test_to_tangent_vec(self, n_reps, atol):
        # TODO: check if it makes sense
        vec = self.data_generator.random_point()
        point = self.data_generator.random_point()

        res = self.space.to_tangent(vec, point)

        vec_data = generate_vectorization_data(
            data=[dict(vector=vec, base_point=point, expected=res, atol=atol)],
            arg_names=["vector", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_to_tangent_is_tangent(self, n_points, atol):
        """Check to_tangent returns tangent vector.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(point)

        expected = gs.ones(n_points, dtype=bool)

        self.test_is_tangent(tangent_vec, point, expected, atol)

    def test_regularize(self, point, expected, atol):
        regularized_point = self.space.regularize(point)
        self.assertAllClose(regularized_point, expected, atol=atol)

    @pytest.mark.random
    def test_regularize_belongs(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        regularized_point = self.space.regularize(point)

        expected = gs.ones(n_points, dtype=bool)

        self.test_belongs(regularized_point, expected, atol)

    @pytest.mark.random
    def test_random_tangent_vec_is_tangent(self, n_points, atol):
        """Check to_tangent returns tangent vector.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = gs.ones(n_points, dtype=bool)
        self.test_is_tangent(tangent_vec, base_point, expected, atol)

    @pytest.mark.shape
    def test_random_tangent_vec_shape(self, n_points):
        """Check random point shape.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        """
        point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(point)

        expected_ndim = self.space.point_ndim + int(n_points > 1)
        self.assertEqual(gs.ndim(point), expected_ndim)

        self.assertAllEqual(
            gs.shape(tangent_vec)[-self.space.point_ndim :], self.space.shape
        )

        if n_points > 1:
            self.assertEqual(gs.shape(tangent_vec)[0], n_points)


class ManifoldTestCase(_ManifoldTestCaseMixins, TestCase):
    pass



"""Common interface to hyperbolic spaces.

Lead author: Thomas Gerald.
"""

import geomstats.errors as errors
from geomstats.geometry._hyperbolic import HyperbolicDiffeo  # noqa
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.poincare_ball import PoincareBall
from geomstats.geometry.poincare_half_space import PoincareHalfSpace


class Hyperbolic:
    """Class for the n-dimensional Hyperbolic space.

    This class is a common interface to the different models of hyperbolic
    geometry:

    - the hyperboloid, embedded in Minkowski space of dimension dim + 1 as the set of
      points whose squared norm is equal to -1. This representation is called
      `extrinsic` here.
    - the Poincare ball, the open ball of the Euclidean space of dimension dim.
    - the Poincare half-space, the open space of points of the Euclidean
      space of  dimension dim, whose last coordinate is positive.

    Parameters
    ----------
    dim : int
        Dimension of the hyperbolic space.
    coords_type : str, {'extrinsic', 'ball', 'half-space'}
        Default coordinates to represent points in hyperbolic space.
        Optional, default: 'extrinsic'.
    """

    def __new__(cls, dim, coords_type="extrinsic", equip=True):
        """Instantiate class that corresponds to the coords_type."""
        errors.check_parameter_accepted_values(
            coords_type,
            "coords_type",
            ["extrinsic", "ball", "half-space"],
        )
        if coords_type == "extrinsic":
            return Hyperboloid(dim, equip=equip)
        if coords_type == "ball":
            return PoincareBall(dim, equip=equip)
        return PoincareHalfSpace(dim, equip=equip)
